@@ -75,9 +75,6 @@ def main():
         **load_config(f'{this_script_dir}/buildconfig/{args.platform}.txt')
     }
 
-    if args.binaries:
-        options_dict = remove_gpl_options(options_dict)
-
     options = build_utils.cmake_options_from_dict(options_dict)
 
     if WINDOWS:
@@ -151,7 +148,7 @@ def main():
     if testing:
         extra_ctest_flags = ""
         if WINDOWS:
-            extra_ctest_flags += "--repeat until-pass:4 "
+            extra_ctest_flags += "--repeat until-pass:5 "
             extra_ctest_flags += "--build-config " + args.buildtype
 
         ctest_returncode = run_ctest(extra_ctest_flags)
@@ -215,13 +212,6 @@ def parse_args():
 def print_trace():
     build_utils.log.print()
 
-def remove_gpl_options(options_dict: dict):
-    gpl_options = ['builtin_fftw3', 'builtin_gsl', 'builtin_unuran', 'fftw3', 'mathmore', 'pythia6', 'pythia8', 'unuran']
-    for opt in gpl_options:
-        options_dict[opt] = 'off'
-    return options_dict
-
-
 @github_log_group("Clean up from previous runs")
 def cleanup_previous_build():
     # runners should never have root permissions but be on the safe side
@@ -241,7 +231,8 @@ def cleanup_previous_build():
     else:
         # mac/linux/POSIX
         result = subprocess_with_log(f"""
-            rm -rf {WORKDIR}/*
+            rm -rf {WORKDIR}
+            mkdir -p {WORKDIR}
         """)
 
     if result != 0:
@@ -375,10 +366,11 @@ def dump_requested_config(options):
 @github_log_group("Build")
 def cmake_build(buildtype):
     generator_flags = "-- '-verbosity:minimal'" if WINDOWS else ""
+    parallel_jobs = "4" if WINDOWS else str(os.cpu_count())
 
     builddir = os.path.join(WORKDIR, "build")
     result = subprocess_with_log(f"""
-        cmake --build '{builddir}' --config '{buildtype}' --parallel '{os.cpu_count()}' {generator_flags}
+        cmake --build '{builddir}' --config '{buildtype}' --parallel '{parallel_jobs}' {generator_flags}
     """)
 
     if result != 0:
