@@ -15,6 +15,33 @@ namespace CPyCppyy {
 class Executor;
 class Converter;
 
+class PyCallArgs {
+public:
+    PyCallArgs(CPPInstance*& self, CPyCppyy_PyArgs_t args, size_t nargsf, PyObject* kwds)
+        : fSelf(self), fArgs(args), fNArgsf(nargsf), fKwds(kwds), fFlags(kNone) {}
+    ~PyCallArgs();
+
+    enum ECleanupFlags {
+        kNone           = 0x0000,
+        kIsOffset       = 0x0001, // args were offset by 1 to drop self
+        kSelfSwap       = 0x0002, // args[-1] and self need swapping
+        kArgsSwap       = 0x0004, // args[0] and args[1] need swapping
+#if PY_VERSION_HEX >= 0x03080000
+        kDoFree         = 0x0008, // args need to be free'd (vector call only)
+        kDoItemDecref   = 0x0010  // items in args need a decref (vector call only)
+#else
+        kDoDecref       = 0x0020  // args need a decref
+#endif
+    };
+
+public:
+    CPPInstance*&     fSelf;
+    CPyCppyy_PyArgs_t fArgs;
+    size_t            fNArgsf;
+    PyObject*         fKwds;
+    int               fFlags;
+};
+
 class CPPMethod : public PyCallable {
 public:
     CPPMethod(Cppyy::TCppScope_t scope, Cppyy::TCppMethod_t method);
@@ -25,14 +52,16 @@ public:
 public:
     virtual PyObject* GetSignature(bool show_formalargs = true);
     virtual PyObject* GetPrototype(bool show_formalargs = true);
+    virtual PyObject* GetTypeName();
     virtual PyObject* Reflex(Cppyy::Reflex::RequestId_t request,
                              Cppyy::Reflex::FormatId_t = Cppyy::Reflex::OPTIMAL);
+
     virtual int       GetPriority();
-    virtual bool IsGreedy();
+    virtual bool      IsGreedy();
 
     virtual int       GetMaxArgs();
     virtual PyObject* GetCoVarNames();
-    virtual PyObject* GetArgDefault(int iarg);
+    virtual PyObject* GetArgDefault(int iarg, bool silent=true);
     virtual bool      IsConst();
 
     virtual PyObject* GetScopeProxy();
@@ -43,21 +72,18 @@ public:
     virtual int       GetArgMatchScore(PyObject* args_tuple);
 
 public:
-    virtual PyObject* Call(
-        CPPInstance*& self, PyObject* args, PyObject* kwds, CallContext* ctxt = nullptr);
+    virtual PyObject* Call(CPPInstance*& self,
+        CPyCppyy_PyArgs_t args, size_t nargsf, PyObject* kwds, CallContext* ctxt = nullptr);
 
 protected:
-    virtual PyObject* PreProcessArgs(CPPInstance*& self, PyObject* args, PyObject* kwds);
+    virtual bool ProcessArgs(PyCallArgs& args);
 
-    bool      Initialize(CallContext* ctxt = nullptr);
-    PyObject* ProcessKeywords(PyObject* self, PyObject* args, PyObject* kwds);
-    bool      ConvertAndSetArgs(PyObject* args, CallContext* ctxt = nullptr);
+    bool Initialize(CallContext* ctxt = nullptr);
+    bool ProcessKwds(PyObject* self_in, PyCallArgs& args);
+    bool ConvertAndSetArgs(CPyCppyy_PyArgs_t, size_t nargsf, CallContext* ctxt = nullptr);
     PyObject* Execute(void* self, ptrdiff_t offset, CallContext* ctxt = nullptr);
 
     Cppyy::TCppMethod_t GetMethod()   { return fMethod; }
-// TODO: the following is a special case to allow shimming of the
-// constructor; there's probably a better way ...
-    void SetMethod(Cppyy::TCppMethod_t m) { fMethod = m; }
     Cppyy::TCppScope_t  GetScope()    { return fScope; }
     Executor*           GetExecutor() { return fExecutor; }
     std::string         GetSignatureString(bool show_formalargs = true);
@@ -68,6 +94,7 @@ protected:
 private:
     void Copy_(const CPPMethod&);
     void Destroy_();
+    bool VerifyArgCount_(Py_ssize_t);
 
     PyObject* ExecuteFast(void*, ptrdiff_t, CallContext*);
     PyObject* ExecuteProtected(void*, ptrdiff_t, CallContext*);
