@@ -665,9 +665,6 @@ PyTypeObject CPPScope_Type = {
     Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE
 #if PY_VERSION_HEX >= 0x03040000
         | Py_TPFLAGS_TYPE_SUBCLASS
-#endif
-#if PY_VERSION_HEX >= 0x03120000
-        | Py_TPFLAGS_MANAGED_DICT | Py_TPFLAGS_HAVE_GC
 #endif
         ,                          // tp_flags
     (char*)"CPyCppyy metatype (internal)",        // tp_doc
