@@ -17,6 +17,7 @@
 #include "Utility.h"
 
 #define CPYCPPYY_INTERNAL 1
+#include "CPyCppyy/DispatchPtr.h"
 namespace CPyCppyy {
 void* Instance_AsVoidPtr(PyObject* pyobject);
 PyObject* Instance_FromVoidPtr(
@@ -29,11 +30,17 @@ PyObject* Instance_FromVoidPtr(
 #include <map>
 #include <set>
 #include <string>
+#include <iostream>
 #include <sstream>
 #include <utility>
 #include <vector>
 
 
+// Note: as of py3.11, dictionary objects no longer carry a function pointer for
+// the lookup, so it can no longer be shimmed and "from cppyy.interactive import *"
+// thus no longer works.
+#if PY_VERSION_HEX < 0x030b0000
+
 //- from Python's dictobject.c -------------------------------------------------
 #if PY_VERSION_HEX >= 0x03030000
     typedef struct PyDictKeyEntry {
@@ -73,6 +80,8 @@ PyObject* Instance_FromVoidPtr(
 
 #endif
 
+#endif // PY_VERSION_HEX < 0x030b0000
+
 //- data -----------------------------------------------------------------------
 static PyObject* nullptr_repr(PyObject*)
 {
@@ -149,6 +158,52 @@ static PyTypeObject PyNullPtr_t_Type = {
 #if PY_VERSION_HEX >= 0x03040000
     , 0                  // tp_finalize
 #endif
+#if PY_VERSION_HEX >= 0x03080000
+    , 0                  // tp_vectorcall
+#endif
+#if PY_VERSION_HEX >= 0x030c0000
+    , 0                  // tp_watched
+#endif
+};
+
+
+static PyObject* default_repr(PyObject*)
+{
+    return CPyCppyy_PyText_FromString("type default");
+}
+
+static void default_dealloc(PyObject*)
+{
+    Py_FatalError("deallocating default");
+}
+
+static PyTypeObject PyDefault_t_Type = {
+    PyVarObject_HEAD_INIT(&PyType_Type, 0)
+    "default_t",         // tp_name
+    sizeof(PyObject),    // tp_basicsize
+    0,                   // tp_itemsize
+    default_dealloc,     // tp_dealloc (never called)
+    0, 0, 0, 0,
+    default_repr,        // tp_repr
+    0, 0, 0,
+    (hashfunc)_Py_HashPointer, // tp_hash
+    0, 0, 0, 0, 0, Py_TPFLAGS_DEFAULT, 0, 0, 0, 0, 0, 0, 0,
+    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
+#if PY_VERSION_HEX >= 0x02030000
+    , 0                  // tp_del
+#endif
+#if PY_VERSION_HEX >= 0x02060000
+    , 0                  // tp_version_tag
+#endif
+#if PY_VERSION_HEX >= 0x03040000
+    , 0                  // tp_finalize
+#endif
+#if PY_VERSION_HEX >= 0x03080000
+    , 0                 // tp_vectorcall
+#endif
+#if PY_VERSION_HEX >= 0x030c0000
+    , 0                 // tp_watched
+#endif
 };
 
 namespace {
@@ -158,7 +213,12 @@ PyObject _CPyCppyy_NullPtrStruct = {
     1, &PyNullPtr_t_Type
 };
 
-// TOOD: refactor with Converters.cxx
+PyObject _CPyCppyy_DefaultStruct = {
+    _PyObject_EXTRA_INIT
+    1, &PyDefault_t_Type
+};
+
+// TODO: refactor with Converters.cxx
 struct CPyCppyy_tagCDataObject {       // non-public (but stable)
     PyObject_HEAD
     char* b_ptr;
@@ -171,12 +231,15 @@ namespace CPyCppyy {
     PyObject* gThisModule    = nullptr;
     PyObject* gPyTypeMap     = nullptr;
     PyObject* gNullPtrObject = nullptr;
+    PyObject* gDefaultObject = nullptr;
     PyObject* gBusException  = nullptr;
     PyObject* gSegvException = nullptr;
     PyObject* gIllException  = nullptr;
     PyObject* gAbrtException = nullptr;
     std::map<std::string, std::vector<PyObject*>> gPythonizations;
     std::set<Cppyy::TCppType_t> gPinnedTypes;
+    std::ostringstream gCapturedError;
+    std::streambuf* gOldErrorBuffer = nullptr;
 }
 
 
@@ -185,7 +248,9 @@ namespace {
 
 using namespace CPyCppyy;
 
+
 //----------------------------------------------------------------------------
+#if PY_VERSION_HEX < 0x030b0000
 namespace {
 
 class GblGetter {
@@ -338,9 +403,12 @@ PyDictEntry* CPyCppyyLookDictString(PyDictObject* mp, PyObject* key, long hash)
     return ep;
 }
 
+#endif // PY_VERSION_HEX < 0x030b0000
+
 //----------------------------------------------------------------------------
 static PyObject* SetCppLazyLookup(PyObject*, PyObject* args)
 {
+#if PY_VERSION_HEX < 0x030b0000
 // Modify the given dictionary to install the lookup function that also
 // tries the global C++ namespace before failing. Called on a module's dictionary,
 // this allows for lazy lookups. This works fine for p3.2 and earlier, but should
@@ -351,6 +419,12 @@ static PyObject* SetCppLazyLookup(PyObject*, PyObject* args)
         return nullptr;
 
     CPYCPPYY_GET_DICT_LOOKUP(dict) = CPyCppyyLookDictString;
+#else
+// As of py3.11, there is no longer a lookup function pointer in the dict object
+// to replace. Since this feature is not widely advertised, it's simply dropped
+    PyErr_Warn(PyExc_RuntimeWarning, (char*)"lazy lookup is no longer supported");
+    (void)args; // avoid warning about unused parameter
+#endif
 
     Py_RETURN_NONE;
 }
@@ -409,13 +483,13 @@ static void* GetCPPInstanceAddress(const char* fname, PyObject* args, PyObject*
                 return nullptr;
             }
 
-            // this is an address of an address (i.e. &myobj, with myobj of type MyObj*)
-            // note that the return result may be null
+        // this is an address of an address (i.e. &myobj, with myobj of type MyObj*)
+        // note that the return result may be null
             if (!byref) return ((CPPInstance*)pyobj)->GetObject();
             return &((CPPInstance*)pyobj)->GetObjectRaw();
 
         } else if (CPyCppyy_PyText_Check(pyobj)) {
-        // special cases for acces to the CPyCppyy API
+        // special cases for access to the CPyCppyy API
             std::string req = CPyCppyy_PyText_AsString((PyObject*)pyobj);
             if (req == "Instance_AsVoidPtr")
                 return (void*)&Instance_AsVoidPtr;
@@ -454,8 +528,8 @@ static PyObject* addressof(PyObject* /* dummy */, PyObject* args, PyObject* kwds
                 return nullptr;
             }
 
-            Cppyy::TCppFuncAddr_t addr = methods[0]->GetFunctionAddress();
-            return PyLong_FromLongLong((intptr_t)addr);
+            Cppyy::TCppFuncAddr_t caddr = methods[0]->GetFunctionAddress();
+            return PyLong_FromLongLong((intptr_t)caddr);
         }
 
     // C functions (incl. ourselves)
@@ -494,7 +568,7 @@ static PyObject* AsCObject(PyObject* /* unused */, PyObject* args, PyObject* kwd
 }
 
 //----------------------------------------------------------------------------
-static PyObject* AsCapsule(PyObject* /* dummy */, PyObject* args, PyObject* kwds)
+static PyObject* AsCapsule(PyObject* /* unused */, PyObject* args, PyObject* kwds)
 {
 // Return object proxy as an opaque PyCapsule.
     void* addr = GetCPPInstanceAddress("as_capsule", args, kwds);
@@ -508,7 +582,7 @@ static PyObject* AsCapsule(PyObject* /* dummy */, PyObject* args, PyObject* kwds
 }
 
 //----------------------------------------------------------------------------
-static PyObject* AsCTypes(PyObject* /* dummy */, PyObject* args, PyObject* kwds)
+static PyObject* AsCTypes(PyObject* /* unused */, PyObject* args, PyObject* kwds)
 {
 // Return object proxy as a ctypes c_void_p
     void* addr = GetCPPInstanceAddress("as_ctypes", args, kwds);
@@ -533,6 +607,43 @@ static PyObject* AsCTypes(PyObject* /* dummy */, PyObject* args, PyObject* kwds)
     return ref;
 }
 
+//----------------------------------------------------------------------------
+static PyObject* AsMemoryView(PyObject* /* unused */, PyObject* pyobject)
+{
+// Return a raw memory view on arrays of PODs.
+    if (!CPPInstance_Check(pyobject)) {
+        PyErr_SetString(PyExc_TypeError, "C++ object proxy expected");
+        return nullptr;
+    }
+
+    CPPInstance* pyobj = (CPPInstance*)pyobject;
+    Cppyy::TCppType_t klass = ((CPPClass*)Py_TYPE(pyobject))->fCppType;
+
+    Py_ssize_t array_len = pyobj->ArrayLength();
+
+    if (array_len < 0 || !Cppyy::IsAggregate(klass)) {
+        PyErr_SetString(
+            PyExc_TypeError, "object is not a proxy to an array of PODs of known size");
+        return nullptr;
+    }
+
+    Py_buffer view;
+
+    view.obj            = pyobject;
+    view.buf            = pyobj->GetObject();
+    view.itemsize       = Cppyy::SizeOf(klass);
+    view.len            = view.itemsize * array_len;
+    view.readonly       = 0;
+    view.format         = NULL;   // i.e. "B" assumed
+    view.ndim           = 1;
+    view.shape          = NULL;
+    view.strides        = NULL;
+    view.suboffsets     = NULL;
+    view.internal       = NULL;
+
+    return PyMemoryView_FromBuffer(&view);
+}
+
 //----------------------------------------------------------------------------
 static PyObject* BindObject(PyObject*, PyObject* args, PyObject* kwds)
 {
@@ -540,20 +651,121 @@ static PyObject* BindObject(PyObject*, PyObject* args, PyObject* kwds)
     Py_ssize_t argc = PyTuple_GET_SIZE(args);
     if (argc != 2) {
         PyErr_Format(PyExc_TypeError,
-            "BindObject takes exactly 2 argumenst (" PY_SSIZE_T_FORMAT " given)", argc);
+            "bind_object takes 2 positional arguments but (" PY_SSIZE_T_FORMAT " were given)", argc);
         return nullptr;
     }
 
-// try to convert first argument: either PyCapsule/CObject or long integer
-    PyObject* pyaddr = PyTuple_GET_ITEM(args, 0);
+// convert 2nd argument first (used for both pointer value and instance cases)
+    Cppyy::TCppType_t cast_type = 0;
+    PyObject* arg1 = PyTuple_GET_ITEM(args, 1);
+    if (!CPyCppyy_PyText_Check(arg1)) {          // not string, then class
+        if (CPPScope_Check(arg1))
+            cast_type = ((CPPClass*)arg1)->fCppType;
+        else
+            arg1 = PyObject_GetAttr(arg1, PyStrings::gName);
+    } else
+        Py_INCREF(arg1);
 
+    if (!cast_type && arg1) {
+        cast_type = (Cppyy::TCppType_t)Cppyy::GetScope(CPyCppyy_PyText_AsString(arg1));
+        Py_DECREF(arg1);
+    }
+
+    if (!cast_type) {
+        PyErr_SetString(PyExc_TypeError,
+            "bind_object expects a valid class or class name as an argument");
+        return nullptr;
+    }
+
+// next, convert the first argument, some pointer value or a pre-existing instance
+    PyObject* arg0 = PyTuple_GET_ITEM(args, 0);
+
+    if (CPPInstance_Check(arg0)) {
+    // if this instance's class has a relation to the requested one, calculate the
+    // offset, erase if from any caches, and update the pointer and type
+        CPPInstance* arg0_pyobj = (CPPInstance*)arg0;
+        Cppyy::TCppType_t cur_type = arg0_pyobj->ObjectIsA(false /* check_smart */);
+
+        bool isPython = CPPScope_Check(arg1) && \
+            (((CPPClass*)arg1)->fFlags & CPPScope::kIsPython);
+
+        if (cur_type == cast_type && !isPython) {
+            Py_INCREF(arg0);      // nothing to do
+            return arg0;
+        }
+
+        int direction = 0;
+        Cppyy::TCppType_t base = 0, derived = 0;
+        if (Cppyy::IsSubtype(cast_type, cur_type)) {
+            derived = cast_type;
+            base    = cur_type;
+            direction = -1;      // down-cast
+        } else if (Cppyy::IsSubtype(cur_type, cast_type)) {
+            base    = cast_type;
+            derived = cur_type;
+            direction =  1;      // up-cast
+        } else {
+            PyErr_SetString(PyExc_TypeError,
+                "provided instance and provided target type are unrelated");
+            return nullptr;
+        }
+
+        Cppyy::TCppObject_t address = (Cppyy::TCppObject_t)arg0_pyobj->GetObject();
+        ptrdiff_t offset = Cppyy::GetBaseOffset(derived, base, address, direction);
+
+    // it's debatable whether a new proxy should be created rather than updating
+    // the old, but changing the old object would be changing the behavior of all
+    // code that has a reference to it, which may not be the intention if the cast
+    // is on a C++ data member; this probably is the "least surprise" option
+
+    // ownership is taken over as needed, again following the principle of "least
+    // surprise" as most likely only the cast object will be retained
+        bool owns = arg0_pyobj->fFlags & CPPInstance::kIsOwner;
+
+        if (!isPython) {
+        // ordinary C++ class
+            PyObject* pyobj = BindCppObjectNoCast(
+                (void*)((intptr_t)address + offset), cast_type, owns ? CPPInstance::kIsOwner : 0);
+            if (owns && pyobj) arg0_pyobj->CppOwns();
+            return pyobj;
+
+        } else {
+        // rebinding to a Python-side class, create a fresh instance first to be able to
+        // perform a lookup of the original dispatch object and if found, return original
+            void* cast_address = (void*)((intptr_t)address + offset);
+            PyObject* pyobj = ((PyTypeObject*)arg1)->tp_new((PyTypeObject*)arg1, nullptr, nullptr);
+            ((CPPInstance*)pyobj)->GetObjectRaw() = cast_address;
+
+            PyObject* dispproxy = CPyCppyy::GetScopeProxy(cast_type);
+            PyObject* res = PyObject_CallMethodOneArg(dispproxy, PyStrings::gDispGet, pyobj);
+            /* Note: the resultant object is borrowed */
+            if (CPPInstance_Check(res) && ((CPPInstance*)res)->GetObject() == cast_address) {
+                ((CPPInstance*)pyobj)->CppOwns();     // make sure C++ object isn't deleted
+                Py_DECREF(pyobj);                     //  on DECREF (is default, but still)
+                pyobj = res;
+            } else {
+                if (res) Py_DECREF(res);              // most likely Py_None
+                else PyErr_Clear();                   // should not happen
+            }
+            Py_DECREF(dispproxy);
+
+            if (pyobj && owns) {
+                arg0_pyobj->CppOwns();
+                ((CPPInstance*)pyobj)->PythonOwns();
+            }
+
+            return pyobj;
+        }
+    }
+
+// not a pre-existing object; get the address and bind
     void* addr = nullptr;
-    if (pyaddr != &_CPyCppyy_NullPtrStruct) {
-        addr = CPyCppyy_PyCapsule_GetPointer(pyaddr, nullptr);
+    if (arg0 != &_CPyCppyy_NullPtrStruct) {
+        addr = CPyCppyy_PyCapsule_GetPointer(arg0, nullptr);
         if (PyErr_Occurred()) {
             PyErr_Clear();
 
-            addr = PyLong_AsVoidPtr(pyaddr);
+            addr = PyLong_AsVoidPtr(arg0);
             if (PyErr_Occurred()) {
                 PyErr_Clear();
 
@@ -561,34 +773,13 @@ static PyObject* BindObject(PyObject*, PyObject* args, PyObject* kwds)
                 Py_ssize_t buflen = Utility::GetBuffer(PyTuple_GetItem(args, 0), '*', 1, addr, false);
                 if (!addr || !buflen) {
                     PyErr_SetString(PyExc_TypeError,
-                        "BindObject requires a CObject or long integer as first argument");
+                        "bind_object requires a CObject/Capsule, long integer, buffer, or instance as first argument");
                     return nullptr;
                 }
             }
         }
     }
 
-    Cppyy::TCppType_t klass = 0;
-    PyObject* pyname = PyTuple_GET_ITEM(args, 1);
-    if (!CPyCppyy_PyText_Check(pyname)) {         // not string, then class
-        if (CPPScope_Check(pyname))
-            klass = ((CPPClass*)pyname)->fCppType;
-        else
-            pyname = PyObject_GetAttr(pyname, PyStrings::gName);
-    } else
-        Py_INCREF(pyname);
-
-    if (!klass && pyname) {
-        klass = (Cppyy::TCppType_t)Cppyy::GetScope(CPyCppyy_PyText_AsString(pyname));
-        Py_DECREF(pyname);
-    }
-
-    if (!klass) {
-        PyErr_SetString(PyExc_TypeError,
-            "BindObject expects a valid class or class name as an argument");
-        return nullptr;
-    }
-
     bool do_cast = false;
     if (kwds) {
         PyObject* cast = PyDict_GetItemString(kwds, "cast");
@@ -596,9 +787,9 @@ static PyObject* BindObject(PyObject*, PyObject* args, PyObject* kwds)
     }
 
     if (do_cast)
-        return BindCppObject(addr, klass);
+        return BindCppObject(addr, cast_type);
 
-    return BindCppObjectNoCast(addr, klass);
+    return BindCppObjectNoCast(addr, cast_type);
 }
 
 //----------------------------------------------------------------------------
@@ -659,6 +850,34 @@ static PyObject* RemovePythonization(PyObject*, PyObject* args)
     Py_RETURN_FALSE;
 }
 
+//----------------------------------------------------------------------------
+static PyObject* PinType(PyObject*, PyObject* pyclass)
+{
+// Add a pinning so that objects of type `derived' are interpreted as
+// objects of type `base'.
+    if (!CPPScope_Check(pyclass)) {
+        PyErr_SetString(PyExc_TypeError, "C++ class expected");
+        return nullptr;
+    }
+
+    gPinnedTypes.insert(((CPPClass*)pyclass)->fCppType);
+
+    Py_RETURN_NONE;
+}
+
+//----------------------------------------------------------------------------
+static PyObject* AddTypeReducer(PyObject*, PyObject* args)
+{
+// Add a type reducer to map type2 to type2 on function returns.
+    const char *reducable, *reduced;
+    if (!PyArg_ParseTuple(args, const_cast<char*>("ss"), &reducable, &reduced))
+        return nullptr;
+
+    Cppyy::AddTypeReducer(reducable, reduced);
+
+    Py_RETURN_NONE;
+}
+
 //----------------------------------------------------------------------------
 static PyObject* SetMemoryPolicy(PyObject*, PyObject* args)
 {
@@ -668,9 +887,11 @@ static PyObject* SetMemoryPolicy(PyObject*, PyObject* args)
     if (!PyArg_ParseTuple(args, const_cast<char*>("O!"), &PyInt_Type, &policy))
         return nullptr;
 
+    long old = (long)CallContext::sMemoryPolicy;
+
     long l = PyInt_AS_LONG(policy);
     if (CallContext::SetMemoryPolicy((CallContext::ECallFlags)l)) {
-        Py_RETURN_NONE;
+        return PyInt_FromLong(old);
     }
 
     PyErr_Format(PyExc_ValueError, "Unknown policy %ld", l);
@@ -721,36 +942,28 @@ static PyObject* AddSmartPtrType(PyObject*, PyObject* args)
 }
 
 //----------------------------------------------------------------------------
-static PyObject* PinType(PyObject*, PyObject* pyclass)
+static PyObject* BeginCaptureStderr(PyObject*, PyObject*)
 {
-// Add a pinning so that objects of type `derived' are interpreted as
-// objects of type `base'.
-    if (!CPPScope_Check(pyclass)) {
-        PyErr_SetString(PyExc_TypeError, "C++ class expected");
-        return nullptr;
-    }
-
-    gPinnedTypes.insert(((CPPClass*)pyclass)->fCppType);
+    gOldErrorBuffer = std::cerr.rdbuf();
+    std::cerr.rdbuf(gCapturedError.rdbuf());
 
     Py_RETURN_NONE;
 }
 
 //----------------------------------------------------------------------------
-static PyObject* Cast(PyObject*, PyObject* args)
+static PyObject* EndCaptureStderr(PyObject*, PyObject*)
 {
-// Cast `obj' to type `type'.
-    CPPInstance* obj = nullptr;
-    CPPClass* type = nullptr;
-    if (!PyArg_ParseTuple(args, const_cast<char*>("O!O!"),
-                          &CPPInstance_Type, &obj,
-                          &CPPScope_Type, &type))
-        return nullptr;
-// TODO: this misses an offset calculation, and reference type must not
-// be cast ...
-    return BindCppObjectNoCast(obj->GetObject(), type->fCppType,
-                               obj->fFlags & CPPInstance::kIsReference);
-}
+// restore old rdbuf and return captured result
+    std::cerr.rdbuf(gOldErrorBuffer);
+    gOldErrorBuffer = nullptr;
 
+    std::string capturedError = std::move(gCapturedError).str();
+
+    gCapturedError.str("");
+    gCapturedError.clear();
+
+    return Py_BuildValue("s", capturedError.c_str());
+}
 } // unnamed namespace
 
 
@@ -772,6 +985,8 @@ static PyMethodDef gCPyCppyyMethods[] = {
       METH_VARARGS | METH_KEYWORDS, (char*)"Retrieve address of proxied object or field in a PyCapsule."},
     {(char*) "as_ctypes", (PyCFunction)AsCTypes,
       METH_VARARGS | METH_KEYWORDS, (char*)"Retrieve address of proxied object or field in a ctypes c_void_p."},
+    {(char*) "as_memoryview", (PyCFunction)AsMemoryView,
+      METH_O, (char*)"Represent an array of objects as raw memory."},
     {(char*)"bind_object", (PyCFunction)BindObject,
       METH_VARARGS | METH_KEYWORDS, (char*) "Create an object of given type, from given address."},
     {(char*) "move", (PyCFunction)Move,
@@ -780,6 +995,10 @@ static PyMethodDef gCPyCppyyMethods[] = {
       METH_VARARGS, (char*)"Add a pythonizor."},
     {(char*) "remove_pythonization", (PyCFunction)RemovePythonization,
       METH_VARARGS, (char*)"Remove a pythonizor."},
+    {(char*) "_pin_type", (PyCFunction)PinType,
+      METH_O, (char*)"Install a type pinning."},
+    {(char*) "_add_type_reducer", (PyCFunction)AddTypeReducer,
+      METH_VARARGS, (char*)"Add a type reducer."},
     {(char*) "SetMemoryPolicy", (PyCFunction)SetMemoryPolicy,
       METH_VARARGS, (char*)"Determines object ownership model."},
     {(char*) "SetGlobalSignalPolicy", (PyCFunction)SetGlobalSignalPolicy,
@@ -788,19 +1007,13 @@ static PyMethodDef gCPyCppyyMethods[] = {
       METH_VARARGS, (char*)"Modify held C++ object ownership."},
     {(char*) "AddSmartPtrType", (PyCFunction)AddSmartPtrType,
       METH_VARARGS, (char*) "Add a smart pointer to the list of known smart pointer types."},
-    {(char*) "_pin_type", (PyCFunction)PinType,
-      METH_O, (char*)"Install a type pinning."},
-    {(char*) "Cast", (PyCFunction)Cast,
-      METH_VARARGS, (char*)"Cast the given object to the given type"},
+    {(char*) "_begin_capture_stderr", (PyCFunction)BeginCaptureStderr,
+      METH_NOARGS, (char*) "Begin capturing stderr to a in memory buffer."},
+    {(char*) "_end_capture_stderr", (PyCFunction)EndCaptureStderr,
+      METH_NOARGS, (char*) "End capturing stderr and returns the captured buffer."},
     {nullptr, nullptr, 0, nullptr}
 };
 
-#define QuoteIdent(ident) #ident
-#define QuoteMacro(macro) QuoteIdent(macro)
-#define LIBCPPYY_NAME "libcppyy" QuoteMacro(PY_MAJOR_VERSION) "_" QuoteMacro(PY_MINOR_VERSION)
-
-#define CONCAT(a, b, c, d) a##b##c##d
-#define LIBCPPYY_INIT_FUNCTION(a, b, c, d) CONCAT(a, b, c, d)
 
 #if PY_VERSION_HEX >= 0x03000000
 struct module_state {
@@ -821,9 +1034,10 @@ static int cpycppyymodule_clear(PyObject* m)
     return 0;
 }
 
+
 static struct PyModuleDef moduledef = {
     PyModuleDef_HEAD_INIT,
-    LIBCPPYY_NAME,
+    "libcppyy",
     nullptr,
     sizeof(struct module_state),
     gCPyCppyyMethods,
@@ -836,10 +1050,10 @@ static struct PyModuleDef moduledef = {
 
 //----------------------------------------------------------------------------
 #define CPYCPPYY_INIT_ERROR return nullptr
-LIBCPPYY_INIT_FUNCTION(extern "C" PyObject* PyInit_libcppyy, PY_MAJOR_VERSION, _, PY_MINOR_VERSION) ()
+extern "C" PyObject* PyInit_libcppyy()
 #else
 #define CPYCPPYY_INIT_ERROR return
-LIBCPPYY_INIT_FUNCTION(extern "C" void initlibcppyy, PY_MAJOR_VERSION, _, PY_MINOR_VERSION) ()
+extern "C" void initlibcppyy()
 #endif
 {
 // Initialization of extension module libcppyy.
@@ -853,7 +1067,8 @@ LIBCPPYY_INIT_FUNCTION(extern "C" void initlibcppyy, PY_MAJOR_VERSION, _, PY_MIN
     PyEval_InitThreads();
 #endif
 
-// prepare for lazyness (the insert is needed to capture the most generic lookup
+#if PY_VERSION_HEX < 0x030b0000
+// prepare for laziness (the insert is needed to capture the most generic lookup
 // function, just in case ...)
     PyObject* dict = PyDict_New();
     PyObject* notstring = PyInt_FromLong(5);
@@ -865,12 +1080,13 @@ LIBCPPYY_INIT_FUNCTION(extern "C" void initlibcppyy, PY_MAJOR_VERSION, _, PY_MIN
     gDictLookupOrg = (dict_lookup_func)((PyDictObject*)dict)->ma_lookup;
 #endif
     Py_DECREF(dict);
+#endif // PY_VERSION_HEX < 0x030b0000
 
 // setup this module
 #if PY_VERSION_HEX >= 0x03000000
     gThisModule = PyModule_Create(&moduledef);
 #else
-    gThisModule = Py_InitModule(const_cast<char*>(LIBCPPYY_NAME), gCPyCppyyMethods);
+    gThisModule = Py_InitModule(const_cast<char*>("libcppyy"), gCPyCppyyMethods);
 #endif
     if (!gThisModule)
         CPYCPPYY_INIT_ERROR;
@@ -910,40 +1126,45 @@ LIBCPPYY_INIT_FUNCTION(extern "C" void initlibcppyy, PY_MAJOR_VERSION, _, PY_MIN
         CPYCPPYY_INIT_ERROR;
 
 // inject custom data types
+#if PY_VERSION_HEX < 0x03000000
     if (!Utility::InitProxy(gThisModule, &RefFloat_Type, "Double"))
         CPYCPPYY_INIT_ERROR;
 
     if (!Utility::InitProxy(gThisModule, &RefInt_Type, "Long"))
         CPYCPPYY_INIT_ERROR;
+#endif
 
     if (!Utility::InitProxy(gThisModule, &CustomInstanceMethod_Type, "InstanceMethod"))
         CPYCPPYY_INIT_ERROR;
 
-    if (!Utility::InitProxy(gThisModule, &TupleOfInstances_Type, "InstancesArray"))
+    if (!Utility::InitProxy(gThisModule, &TupleOfInstances_Type, "InstanceArray"))
        CPYCPPYY_INIT_ERROR;
 
-    if (!Utility::InitProxy(gThisModule, &InstanceArrayIter_Type, "instancearrayiter"))
-       CPYCPPYY_INIT_ERROR;
+    if (!Utility::InitProxy(gThisModule, &LowLevelView_Type, "LowLevelView"))
+        CPYCPPYY_INIT_ERROR;
 
     if (!Utility::InitProxy(gThisModule, &PyNullPtr_t_Type, "nullptr_t"))
         CPYCPPYY_INIT_ERROR;
 
-// initialize low level ptr type, but do not inject in gThisModule
-    if (PyType_Ready(&LowLevelView_Type) < 0)
+// custom iterators
+    if (PyType_Ready(&InstanceArrayIter_Type) < 0)
         CPYCPPYY_INIT_ERROR;
 
-// custom iterators
     if (PyType_Ready(&IndexIter_Type) < 0)
         CPYCPPYY_INIT_ERROR;
 
     if (PyType_Ready(&VectorIter_Type) < 0)
         CPYCPPYY_INIT_ERROR;
 
-// inject identifiable nullptr
+// inject identifiable nullptr and default
     gNullPtrObject = (PyObject*)&_CPyCppyy_NullPtrStruct;
     Py_INCREF(gNullPtrObject);
     PyModule_AddObject(gThisModule, (char*)"nullptr", gNullPtrObject);
 
+    gDefaultObject = (PyObject*)&_CPyCppyy_DefaultStruct;
+    Py_INCREF(gDefaultObject);
+    PyModule_AddObject(gThisModule, (char*)"default", gDefaultObject);
+
 // C++-specific exceptions
     PyObject* cppfatal = PyErr_NewException((char*)"cppyy.ll.FatalError", nullptr, nullptr);
     PyModule_AddObject(gThisModule, (char*)"FatalError", cppfatal);
