@@ -12,7 +12,7 @@
 #pragma link off all classes;
 #pragma link off all functions;
 
-#pragma link C++ class TGeoVGConverter+;
-#pragma link C++ class TGeoVGShape+;
+#pragma link C++ class TGeoVGConverter + ;
+#pragma link C++ class TGeoVGShape - ;
 
 #endif
