@@ -13,6 +13,8 @@
 #include "ROOT/REveShape.hxx"
 #include "ROOT/REveRenderData.hxx"
 #include "ROOT/REveRGBAPalette.hxx"
+#include "ROOT/REveManager.hxx"
+#include "ROOT/REveTrans.hxx"
 
 #include "TRandom.h"
 #include <cassert>
@@ -52,9 +54,6 @@ REveBoxSet::REveBoxSet(const char* n, const char* t) :
    REveDigitSet  (n, t),
 
    fBoxType      (kBT_Undef),
-   fDefWidth     (1),
-   fDefHeight    (1),
-   fDefDepth     (1),
 
    fBoxSkip      (0),
 
@@ -74,14 +73,12 @@ Int_t REveBoxSet::SizeofAtom(REveBoxSet::EBoxType_e bt)
    static const REveException eH("REveBoxSet::SizeofAtom ");
 
    switch (bt) {
-      case kBT_Undef:                return 0;
-      case kBT_FreeBox:              return sizeof(BFreeBox_t);
-      case kBT_AABox:                return sizeof(BAABox_t);
-      case kBT_AABoxFixedDim:        return sizeof(BAABoxFixedDim_t);
-      case kBT_Cone:                 return sizeof(BCone_t);
-      case kBT_EllipticCone:         return sizeof(BEllipticCone_t);
-      case kBT_Hex:                  return sizeof(BHex_t);
-      default:                       throw(eH + "unexpected atom type.");
+      case kBT_Undef:                  return 0;
+      case kBT_FreeBox:                return sizeof(BFreeBox_t);
+      case kBT_Instanced:              return sizeof(Instanced_t);
+      case kBT_InstancedScaled:         return sizeof(InstancedScaled_t);
+      case kBT_InstancedScaledRotated:  return sizeof(InstancedScaledRotated_t);
+      default:                        throw(eH + "unexpected atom type.");
    }
    return 0;
 }
@@ -113,7 +110,7 @@ void REveBoxSet::Reset()
 /// Create a new box from a set of 8 vertices.
 /// To be used for box-type kBT_FreeBox.
 
-void REveBoxSet::AddBox(const Float_t* verts)
+void REveBoxSet::AddFreeBox(const Float_t* verts)
 {
    static const REveException eH("REveBoxSet::AddBox ");
 
@@ -130,15 +127,15 @@ void REveBoxSet::AddBox(const Float_t* verts)
 /// specified dimensions.
 /// To be used for box-type kBT_AABox.
 
-void REveBoxSet::AddBox(Float_t a, Float_t b, Float_t c, Float_t w, Float_t h, Float_t d)
+void REveBoxSet::AddInstanceScaled(Float_t a, Float_t b, Float_t c, Float_t w, Float_t h, Float_t d)
 {
    static const REveException eH("REveBoxSet::AddBox ");
 
-   if (fBoxType != kBT_AABox)
+   if (fBoxType != kBT_InstancedScaled)
       throw(eH + "expect axis-aligned box-type.");
 
-   BAABox_t* box = (BAABox_t*) NewDigit();
-   box->fA = a; box->fB = b; box->fC = c;
+   InstancedScaled_t* box = (InstancedScaled_t*) NewDigit();
+   box->fX = a; box->fY = b; box->fZ = c;
    box->fW = w; box->fH = h; box->fD = d;
 }
 
@@ -146,15 +143,28 @@ void REveBoxSet::AddBox(Float_t a, Float_t b, Float_t c, Float_t w, Float_t h, F
 /// Create a new axis-aligned box from at a given position.
 /// To be used for box-type kBT_AABoxFixedDim.
 
-void REveBoxSet::AddBox(Float_t a, Float_t b, Float_t c)
+void REveBoxSet::AddInstance(Float_t a, Float_t b, Float_t c)
 {
    static const REveException eH("REveBoxSet::AddBox ");
 
-   if (fBoxType != kBT_AABoxFixedDim)
+   if (fBoxType != kBT_Instanced)
       throw(eH + "expect axis-aligned fixed-dimension box-type.");
 
-   BAABoxFixedDim_t* box = (BAABoxFixedDim_t*) NewDigit();
-   box->fA = a; box->fB = b; box->fC = c;
+   Instanced_t* box = (Instanced_t*) NewDigit();
+   box->fX = a; box->fY = b; box->fZ = c;
+}
+
+////////////////////////////////////////////////////////////////////////////////
+/// Create shape with arbitrary transformtaion
+///
+void REveBoxSet::AddInstanceMat4(const Float_t* arr)
+{
+   static const REveException eH("REveBoxSet::AddMat4Box ");
+   if (fBoxType != kBT_InstancedScaledRotated)
+      throw(eH + "expect Mat4 box-type.");
+
+   InstancedScaledRotated_t* b = (InstancedScaledRotated_t*) NewDigit();
+   memcpy(b->fMat, arr, sizeof(b->fMat));
 }
 
 ////////////////////////////////////////////////////////////////////////////////
@@ -164,14 +174,23 @@ void REveBoxSet::AddBox(Float_t a, Float_t b, Float_t c)
 void REveBoxSet::AddCone(const REveVector& pos, const REveVector& dir, Float_t r)
 {
    static const REveException eH("REveBoxSet::AddCone ");
+   using namespace TMath;
+   fShapeType = kCone;
 
-   if (fBoxType != kBT_Cone)
-      throw(eH + "expect cone box-type.");
+   REveTrans t;
+   float  h = dir.Mag();
+   float phi   = ATan2(dir.fY, dir.fX);
+   float theta = ATan (dir.fZ / Sqrt(dir.fX*dir.fX + dir.fY*dir.fY));
 
-   BCone_t* cone = (BCone_t*) NewDigit();
-   cone->fPos = pos;
-   cone->fDir = dir;
-   cone->fR   = r;
+   theta =  Pi()/2 -theta;
+   t.RotateLF(1, 2, phi);
+   t.RotateLF(3, 1, theta);
+   t.SetScale(r, r, h);
+   t.SetPos(pos.fX, pos.fY, pos.fZ);
+
+   InstancedScaledRotated_t* cone = (InstancedScaledRotated_t*) NewDigit();
+   for(Int_t i=0; i<16; ++i)
+   cone->fMat[i] = t[i];
 }
 
 ////////////////////////////////////////////////////////////////////////////////
@@ -182,16 +201,23 @@ void REveBoxSet::AddEllipticCone(const REveVector& pos, const REveVector& dir,
                                  Float_t r, Float_t r2, Float_t angle)
 {
    static const REveException eH("REveBoxSet::AddEllipticCone ");
+   using namespace TMath;
+   fShapeType = kCone;
+   REveTrans t;
+   float  h = dir.Mag();
+   float phi   = ATan2(dir.fY, dir.fX);
+   float theta = ATan (dir.fZ / Sqrt(dir.fX*dir.fX + dir.fY*dir.fY));
 
-   if (fBoxType != kBT_EllipticCone)
-      throw(eH + "expect elliptic-cone box-type.");
+   theta =  Pi()/2 -theta;
+   t.RotateLF(1, 2, phi);
+   t.RotateLF(3, 1, theta);
+   t.RotateLF(1, 2, angle * TMath::DegToRad());
+   t.SetScale(r, r2, h);
+   t.SetPos(pos.fX, pos.fY, pos.fZ);
 
-   BEllipticCone_t* cone = (BEllipticCone_t*) NewDigit();
-   cone->fPos = pos;
-   cone->fDir = dir;
-   cone->fR   = r;
-   cone->fR2  = r2;
-   cone->fAngle = angle;
+   InstancedScaledRotated_t* cone = (InstancedScaledRotated_t*) NewDigit();
+   for(Int_t i=0; i<16; ++i)
+   cone->fMat[i] = t[i];
 }
 
 ////////////////////////////////////////////////////////////////////////////////
@@ -201,18 +227,21 @@ void REveBoxSet::AddEllipticCone(const REveVector& pos, const REveVector& dir,
 
 void REveBoxSet::AddHex(const REveVector& pos, Float_t r, Float_t angle, Float_t depth)
 {
-   static const REveException eH("REveBoxSet::AddEllipticCone ");
+   static const REveException eH("REveBoxSet::AddHex ");
 
-   if (fBoxType != kBT_Hex)
+   if (fBoxType != kBT_InstancedScaledRotated)
       throw(eH + "expect hex box-type.");
 
-   BHex_t* hex = (BHex_t*) NewDigit();
-   hex->fPos   = pos;
-   hex->fR     = r;
-   hex->fAngle = angle;
-   hex->fDepth = depth;
-}
+   fShapeType = kHex; 
 
+   InstancedScaledRotated_t* hex = (InstancedScaledRotated_t*) NewDigit();
+   REveTrans t; // AMT do we need to reuse ???
+   t.SetPos(pos.fX, pos.fY, pos.fZ);
+   t.SetScale(r,r,depth);
+   t.RotatePF(1, 2, angle);
+   for(Int_t i=0; i<16; ++i)
+   hex->fMat[i] = t[i];
+}
 ////////////////////////////////////////////////////////////////////////////////
 /// Fill bounding-box information of the base-class TAttBBox (virtual method).
 /// If member 'REveFrameBox* fFrame' is set, frame's corners are used as bbox.
@@ -221,7 +250,7 @@ void REveBoxSet::ComputeBBox()
 {
    static const REveException eH("REveBoxSet::ComputeBBox ");
 
-   if (fFrame != 0)
+   if (fFrame != nullptr)
    {
       BBoxInit();
       Int_t    n    = fFrame->GetFrameSize() / 3;
@@ -253,71 +282,32 @@ void REveBoxSet::ComputeBBox()
          break;
       }
 
-      case kBT_AABox:
+      case kBT_InstancedScaled:
       {
          while (bi.next()) {
-            BAABox_t& b = * (BAABox_t*) bi();
-            BBoxCheckPoint(b.fA, b.fB, b.fC);
-            BBoxCheckPoint(b.fA + b.fW, b.fB + b.fH , b.fC + b.fD);
+            InstancedScaled_t& b = * (InstancedScaled_t*) bi();
+            BBoxCheckPoint(b.fX, b.fY, b.fZ);
+            BBoxCheckPoint(b.fX + b.fW, b.fY + b.fH , b.fZ + b.fD);
          }
          break;
       }
 
-      case kBT_AABoxFixedDim:
+      case kBT_Instanced:
       {
          while (bi.next()) {
-            BAABoxFixedDim_t& b = * (BAABoxFixedDim_t*) bi();
-            BBoxCheckPoint(b.fA, b.fB, b.fC);
-            BBoxCheckPoint(b.fA + fDefWidth, b.fB + fDefHeight , b.fC + fDefDepth);
+            Instanced_t& b = * (Instanced_t*) bi();
+            BBoxCheckPoint(b.fX, b.fY, b.fZ);
+            BBoxCheckPoint(b.fX + fDefWidth, b.fY + fDefHeight , b.fZ + fDefDepth);
          }
          break;
       }
 
-      case kBT_Cone:
-      {
-         Float_t mag2=0, mag2Max=0, rMax=0;
-         while (bi.next()) {
-            BCone_t& b = * (BCone_t*) bi();
-            BBoxCheckPoint(b.fPos.fX, b.fPos.fY, b.fPos.fZ);
-            mag2 = b.fDir.Mag2();
-            if (mag2>mag2Max) mag2Max=mag2;
-            if (b.fR>rMax)    rMax=b.fR;
-         }
-         Float_t off = TMath::Sqrt(mag2Max + rMax*rMax);
-         fBBox[0] -= off;fBBox[2] -= off;fBBox[4] -= off;
-         fBBox[1] += off;fBBox[3] += off;fBBox[5] += off;
-         break;
-      }
-
-      case kBT_EllipticCone:
-      {
-         Float_t mag2=0, mag2Max=0, rMax=0;
-         while (bi.next()) {
-            BEllipticCone_t& b = * (BEllipticCone_t*) bi();
-            BBoxCheckPoint(b.fPos.fX, b.fPos.fY, b.fPos.fZ);
-            mag2 = b.fDir.Mag2();
-            if (mag2>mag2Max) mag2Max=mag2;
-            if (b.fR  > rMax) rMax = b.fR;
-            if (b.fR2 > rMax) rMax = b.fR2;
-         }
-         Float_t off = TMath::Sqrt(mag2Max + rMax*rMax);
-         fBBox[0] -= off;fBBox[2] -= off;fBBox[4] -= off;
-         fBBox[1] += off;fBBox[3] += off;fBBox[5] += off;
-         break;
-      }
-
-      case kBT_Hex:
+      case kBT_InstancedScaledRotated:
       {
          while (bi.next()) {
-            BHex_t& h = * (BHex_t*) bi();
-            BBoxCheckPoint(h.fPos.fX - h.fR, h.fPos.fY - h.fR, h.fPos.fZ);
-            BBoxCheckPoint(h.fPos.fX + h.fR, h.fPos.fY - h.fR, h.fPos.fZ);
-            BBoxCheckPoint(h.fPos.fX + h.fR, h.fPos.fY + h.fR, h.fPos.fZ);
-            BBoxCheckPoint(h.fPos.fX - h.fR, h.fPos.fY + h.fR, h.fPos.fZ);
-            BBoxCheckPoint(h.fPos.fX - h.fR, h.fPos.fY - h.fR, h.fPos.fZ + h.fDepth);
-            BBoxCheckPoint(h.fPos.fX + h.fR, h.fPos.fY - h.fR, h.fPos.fZ + h.fDepth);
-            BBoxCheckPoint(h.fPos.fX + h.fR, h.fPos.fY + h.fR, h.fPos.fZ + h.fDepth);
-            BBoxCheckPoint(h.fPos.fX - h.fR, h.fPos.fY + h.fR, h.fPos.fZ + h.fDepth);
+            InstancedScaledRotated_t& b = * (InstancedScaledRotated_t*) bi();
+            float* a = b.fMat;
+            BBoxCheckPoint(a[12], a[13], a[14]);
          }
          break;
       }
@@ -335,9 +325,54 @@ void REveBoxSet::ComputeBBox()
 
 Int_t REveBoxSet::WriteCoreJson(nlohmann::json &j, Int_t rnr_offset)
 {
-   Int_t ret = REveDigitSet::WriteCoreJson(j, rnr_offset);
    j["boxType"] = int(fBoxType);
+   j["shapeType"] = int(fShapeType);
+   if (fShapeType == kCone)
+   {
+      j["coneCap"] = fDrawConeCap;
+   }
 
+
+   j["instanced"] = Instanced();
+   if (Instanced())
+   {
+      int  N = fPlex.N();
+      int  N_tex = 0;
+      std::string instanceFlag;
+      switch (fBoxType)
+      {
+         case kBT_Instanced:
+            instanceFlag = "FixedDimension";
+            N_tex = N;
+            break;
+         case kBT_InstancedScaled:
+            instanceFlag = "ScalePerDigit";
+            N_tex = 2*N;
+            break;
+         case kBT_InstancedScaledRotated:
+           instanceFlag = "Mat4Trans";
+           N_tex = 4*N;
+           break;
+         default:
+           std::cout << "Ereor:: Unhandled instanced case\n";
+      }
+
+      REveRenderData::CalcTextureSize(N_tex, 4, fTexX, fTexY);
+
+      j["N"] = N;
+      j["texX"] = fTexX;
+      j["texY"] = fTexY;
+      j["instanceFlag"] = instanceFlag;
+      j["defWidth"] = fDefWidth;
+      j["defHeight"] = fDefHeight;
+      j["defDepth"] = fDefDepth;
+
+      // std::cout << "TEXTURE SIZE X " << fTexX << "\n";
+   }
+
+   // AMT:: the base class WroteCoreJson needs to be called after
+   // setting the texture value
+   Int_t ret = REveDigitSet::WriteCoreJson(j, rnr_offset);
    return ret;
 }
 
@@ -357,20 +392,55 @@ void REveBoxSet::BuildRenderData()
 
             if (fValueIsColor) {
                fRenderData->PushI(int(b->fValue));
-            } else if (fSingleColor == false) {
+            } else {
                UChar_t c[4] = {0, 0, 0, 0};
                fPalette->ColorFromValue(b->fValue, fDefaultValue, c);
 
                int value = c[0] + c[1] * 256 + c[2] * 256 * 256;
-
                // printf("box val [%d] values (%d, %d, %d) -> int <%d>\n", b.fValue, c[0], c[1], c[2],  value);
                fRenderData->PushI(value);
             }
          }
       }
    }
+   if (Instanced()) {
+      // printf(" >>> resize render data %d \n", 4 * fTexX * fTexY);
+      fRenderData->ResizeV(4 * fTexX * fTexY);
+   }
 }
 
+////////////////////////////////////////////////////////////////////////////////
+/// Get int value for color. Used for case of instancing.
+///
+unsigned int REveBoxSet::GetColorFromDigit(REveDigitSet::DigitBase_t &digi)
+{
+   if (fSingleColor == false) {
+      if (fValueIsColor) {
+        UChar_t* c = (UChar_t*) & digi.fValue;
+        return (c[0] << 16) + (c[1] << 8) + c[2];
+      } else {
+         // printf("palette\n");
+         UChar_t c[4] = {0, 0, 0, 0};
+         fPalette->ColorFromValue(digi.fValue, fDefaultValue, c);
+         return (c[0] << 16) + (c[1] << 8) + c[2];
+      }
+   }
+   // printf("main color %d\n", GetMainColor());
+   UChar_t c[4] = {0, 0, 0, 0};
+   REveUtil::ColorFromIdx(GetMainColor(), c);
+   // printf("rgb %d %d %d\n", c[0], c[1], c[2]);
+   return (c[0] << 16) + (c[1] << 8) + c[2]; // AMT perhaps this can be ignored
+}
+
+float REveBoxSet::GetColorFromDigitAsFloat(REveDigitSet::DigitBase_t &digit)
+{
+   uint32_t c = GetColorFromDigit(digit);
+   // this line required to avoid strict-aliasing rules warning
+   auto pc = (float *) &c;
+   return *pc;
+}
+
+
 ////////////////////////////////////////////////////////////////////////////////
 /// Write shape data for different cases
 ///
@@ -387,21 +457,32 @@ void REveBoxSet::WriteShapeData(REveDigitSet::DigitBase_t &digit)
       break;
    }
 
-   case REveBoxSet::kBT_AABox: {
-      REveBoxSet::BAABox_t &b = (REveBoxSet::BAABox_t &)(digit);
+   case REveBoxSet::kBT_InstancedScaled: {
+      InstancedScaled_t &b = (InstancedScaled_t &)(digit);
       // position
-      fRenderData->PushV(b.fA, b.fB, b.fC);
-      // dimensions
+      fRenderData->PushV(b.fX, b.fY, b.fZ);
+      fRenderData->PushV(GetColorFromDigitAsFloat(b)); // color ?
       fRenderData->PushV(b.fW, b.fH, b.fD);
-
+      fRenderData->PushV(2.f); // trasp ?
       break;
    }
-
-    case REveBoxSet::kBT_Hex: {
-      REveBoxSet::BHex_t  &b = (REveBoxSet::BHex_t &)(digit);
-      fRenderData->PushV(b.fPos);
-      fRenderData->PushV(b.fR, b.fAngle, b.fDepth);
-
+   case REveBoxSet::kBT_Instanced: {
+      Instanced_t &b =(Instanced_t &)(digit);
+      // position
+      fRenderData->PushV(b.fX, b.fY, b.fZ);
+      fRenderData->PushV(GetColorFromDigitAsFloat(b)); // color ?
+      fRenderData->PushV(2.f); // trasp ?
+      break;
+   }
+   case REveBoxSet::kBT_InstancedScaledRotated: {
+      InstancedScaledRotated_t &b = (InstancedScaledRotated_t &)(digit);
+      float* a = b.fMat;
+      fRenderData->PushV(a[12], a[13], a[14]);
+      fRenderData->PushV(GetColorFromDigitAsFloat(b));
+      // write the first three columns
+      for (int i = 0; i < 12; i++) {
+         fRenderData->PushV(a[i]);
+      }
       break;
    }
 
@@ -414,13 +495,13 @@ void REveBoxSet::WriteShapeData(REveDigitSet::DigitBase_t &digit)
 
 void REveBoxSet::Test(Int_t nboxes)
 {
-   Reset(kBT_AABox, kTRUE, nboxes);
+   Reset(kBT_InstancedScaled, kTRUE, nboxes);
    TRandom rnd(0);
    const Float_t origin = 10, size = 2;
    Int_t color;
    for(Int_t i=0; i<nboxes; ++i)
    {
-      AddBox(origin * rnd.Uniform(-1, 1),
+      AddInstanceScaled(origin * rnd.Uniform(-1, 1),
              origin * rnd.Uniform(-1, 1),
              origin * rnd.Uniform(-1, 1),
              size   * rnd.Uniform(0.1, 1),
@@ -431,3 +512,23 @@ void REveBoxSet::Test(Int_t nboxes)
       DigitValue(color);
    }
 }
+
+////////////////////////////////////////////////////////////////////////////////
+/// Use instancing in RenderCore.
+
+bool REveBoxSet::Instanced()
+{
+   return gEve->IsRCore() && (fBoxType != kBT_FreeBox);
+}
+
+
+/*
+////////////////////////////////////////////////////////////////////////////////
+/// Set DigitShape
+
+bool REveBoxSet::SetDigitShape(const std::vector<float>& vrtBuff, const std::vector<int>& idxBuff)
+{
+   fGeoShape.v = v;
+
+}
+*/
