@@ -74,7 +74,11 @@ private:
    using TH1::IntegralAndError;
 
 public:
-   virtual ~TH3();
+   ~TH3() override;
+           void     AddBinContent(Int_t bin) override;
+           void     AddBinContent(Int_t bin, Double_t w) override;
+   virtual void     AddBinContent(Int_t binx, Int_t biny, Int_t binz);
+   virtual void     AddBinContent(Int_t binx, Int_t biny, Int_t binz, Double_t w);
            Int_t    BufferEmpty(Int_t action = 0) override;
            void     Copy(TObject &hnew) const override;
    virtual Int_t    Fill(Double_t x, Double_t y, Double_t z);
@@ -163,10 +167,12 @@ public:
                                           ,Int_t nbinsy,const Double_t *ybins
                                           ,Int_t nbinsz,const Double_t *zbins);
    TH3C(const TH3C &h3c);
-   virtual ~TH3C();
+   ~TH3C() override;
 
            void      AddBinContent(Int_t bin) override;
            void      AddBinContent(Int_t bin, Double_t w) override;
+           void      AddBinContent(Int_t binx, Int_t biny, Int_t binz) override { AddBinContent(GetBin(binx, biny, binz)); }
+           void      AddBinContent(Int_t binx, Int_t biny, Int_t binz, Double_t w) override { AddBinContent(GetBin(binx, biny, binz), w); }
            void      Copy(TObject &hnew) const override;
            void      Reset(Option_t *option="") override;
            void      SetBinsLength(Int_t n=-1) override;
@@ -201,10 +207,12 @@ public:
                                           ,Int_t nbinsy,const Double_t *ybins
                                           ,Int_t nbinsz,const Double_t *zbins);
    TH3S(const TH3S &h3s);
-   virtual ~TH3S();
+   ~TH3S() override;
 
            void      AddBinContent(Int_t bin) override;
            void      AddBinContent(Int_t bin, Double_t w) override;
+           void      AddBinContent(Int_t binx, Int_t biny, Int_t binz) override { AddBinContent(GetBin(binx, biny, binz)); }
+           void      AddBinContent(Int_t binx, Int_t biny, Int_t binz, Double_t w) override { AddBinContent(GetBin(binx, biny, binz), w); }
            void      Copy(TObject &hnew) const override;
            void      Reset(Option_t *option="") override;
            void      SetBinsLength(Int_t n=-1) override;
@@ -239,10 +247,12 @@ public:
                                           ,Int_t nbinsy,const Double_t *ybins
                                           ,Int_t nbinsz,const Double_t *zbins);
    TH3I(const TH3I &h3i);
-   virtual ~TH3I();
+   ~TH3I() override;
 
            void      AddBinContent(Int_t bin) override;
            void      AddBinContent(Int_t bin, Double_t w) override;
+           void      AddBinContent(Int_t binx, Int_t biny, Int_t binz) override { AddBinContent(GetBin(binx, biny, binz)); }
+           void      AddBinContent(Int_t binx, Int_t biny, Int_t binz, Double_t w) override { AddBinContent(GetBin(binx, biny, binz), w); }
            void      Copy(TObject &hnew) const override;
            void      Reset(Option_t *option="") override;
            void      SetBinsLength(Int_t n=-1) override;
@@ -259,7 +269,46 @@ protected:
            Double_t RetrieveBinContent(Int_t bin) const override { return Double_t (fArray[bin]); }
            void     UpdateBinContent(Int_t bin, Double_t content) override { fArray[bin] = Int_t (content); }
 
-   ClassDefOverride(TH3I,4)  //3-Dim histograms (one 32 bits integer per channel)
+   ClassDefOverride(TH3I,4)  //3-Dim histograms (one 32 bit integer per channel)
+};
+
+
+//________________________________________________________________________
+
+class TH3L : public TH3, public TArrayL64 {
+public:
+   TH3L();
+   TH3L(const char *name,const char *title,Int_t nbinsx,Double_t xlow,Double_t xup
+                                  ,Int_t nbinsy,Double_t ylow,Double_t yup
+                                  ,Int_t nbinsz,Double_t zlow,Double_t zup);
+   TH3L(const char *name,const char *title,Int_t nbinsx,const Float_t *xbins
+                                          ,Int_t nbinsy,const Float_t *ybins
+                                          ,Int_t nbinsz,const Float_t *zbins);
+   TH3L(const char *name,const char *title,Int_t nbinsx,const Double_t *xbins
+                                          ,Int_t nbinsy,const Double_t *ybins
+                                          ,Int_t nbinsz,const Double_t *zbins);
+   TH3L(const TH3L &h3l);
+   ~TH3L() override;
+   void      AddBinContent(Int_t bin) override;
+   void      AddBinContent(Int_t bin, Double_t w) override;
+   void      AddBinContent(Int_t binx, Int_t biny, Int_t binz) override { AddBinContent(GetBin(binx, biny, binz)); }
+   void      AddBinContent(Int_t binx, Int_t biny, Int_t binz, Double_t w) override { AddBinContent(GetBin(binx, biny, binz), w); }
+   void      Copy(TObject &hnew) const override;
+   void      Reset(Option_t *option="") override;
+   void      SetBinsLength(Int_t n=-1) override;
+           TH3L&     operator=(const TH3L &h1);
+   friend  TH3L      operator*(Float_t c1, TH3L &h1);
+   friend  TH3L      operator*(TH3L &h1, Float_t c1) {return operator*(c1,h1);}
+   friend  TH3L      operator+(TH3L &h1, TH3L &h2);
+   friend  TH3L      operator-(TH3L &h1, TH3L &h2);
+   friend  TH3L      operator*(TH3L &h1, TH3L &h2);
+   friend  TH3L      operator/(TH3L &h1, TH3L &h2);
+
+protected:
+   Double_t RetrieveBinContent(Int_t bin) const override { return Double_t (fArray[bin]); }
+   void     UpdateBinContent(Int_t bin, Double_t content) override { fArray[bin] = Int_t (content); }
+
+   ClassDefOverride(TH3L,0)  //3-Dim histograms (one 64 bit integer per channel)
 };
 
 
@@ -278,11 +327,18 @@ public:
                                           ,Int_t nbinsy,const Double_t *ybins
                                           ,Int_t nbinsz,const Double_t *zbins);
    TH3F(const TH3F &h3f);
-   virtual ~TH3F();
+   ~TH3F() override;
 
+           /// Increment bin content by 1.
+           /// Passing an out-of-range bin leads to undefined behavior
            void      AddBinContent(Int_t bin) override {++fArray[bin];}
+           /// Increment bin content by a weight w.
+           /// \warning The value of w is cast to `Float_t` before being added.
+           /// Passing an out-of-range bin leads to undefined behavior
            void      AddBinContent(Int_t bin, Double_t w) override
                                  {fArray[bin] += Float_t (w);}
+           void      AddBinContent(Int_t binx, Int_t biny, Int_t binz) override { AddBinContent(GetBin(binx, biny, binz)); }
+           void      AddBinContent(Int_t binx, Int_t biny, Int_t binz, Double_t w) override { AddBinContent(GetBin(binx, biny, binz), w); }
            void      Copy(TObject &hnew) const override;
            void      Reset(Option_t *option="") override;
            void      SetBinsLength(Int_t n=-1) override;
@@ -317,11 +373,17 @@ public:
                                           ,Int_t nbinsy,const Double_t *ybins
                                           ,Int_t nbinsz,const Double_t *zbins);
    TH3D(const TH3D &h3d);
-   virtual ~TH3D();
+   ~TH3D() override;
 
+           /// Increment bin content by 1.
+           /// Passing an out-of-range bin leads to undefined behavior
            void      AddBinContent(Int_t bin) override {++fArray[bin];}
+           /// Increment bin content by a weight w
+           /// Passing an out-of-range bin leads to undefined behavior
            void      AddBinContent(Int_t bin, Double_t w) override
                                  {fArray[bin] += Double_t (w);}
+           void      AddBinContent(Int_t binx, Int_t biny, Int_t binz) override { AddBinContent(GetBin(binx, biny, binz)); }
+           void      AddBinContent(Int_t binx, Int_t biny, Int_t binz, Double_t w) override { AddBinContent(GetBin(binx, biny, binz), w); }
            void      Copy(TObject &hnew) const override;
            void      Reset(Option_t *option="") override;
            void      SetBinsLength(Int_t n=-1) override;
