@@ -26,11 +26,13 @@
 #include <llvm/ExecutionEngine/Orc/RTDyldObjectLinkingLayer.h>
 #include <llvm/ExecutionEngine/SectionMemoryManager.h>
 #include <llvm/IR/LLVMContext.h>
+#include <llvm/MC/TargetRegistry.h>
 #include <llvm/Support/raw_ostream.h>
 #include <llvm/Support/Host.h>
-#include <llvm/Support/TargetRegistry.h>
 #include <llvm/Target/TargetMachine.h>
 
+#include <optional>
+
 #ifdef __linux__
 #include <sys/stat.h>
 #endif
@@ -39,7 +41,7 @@ using namespace llvm;
 using namespace llvm::jitlink;
 using namespace llvm::orc;
 
-namespace cling {
+namespace {
 
   class ClingMMapper final : public SectionMemoryManager::MemoryMapper {
   public:
@@ -68,10 +70,6 @@ namespace cling {
     }
   };
 
-}
-
-namespace {
-
   // A memory manager for Cling that reserves memory for code and data sections
   // to keep them contiguous for the emission of one module. This is required
   // for working exception handling support since one .eh_frame section will
@@ -133,7 +131,7 @@ namespace {
     AllocInfo m_RWData;
 
   public:
-    ClingMemoryManager(cling::ClingMMapper& MMapper) : Super(&MMapper) {}
+    ClingMemoryManager(ClingMMapper& MMapper) : Super(&MMapper) {}
 
     uint8_t* allocateCodeSection(uintptr_t Size, unsigned Alignment,
                                  unsigned SectionID,
@@ -169,23 +167,23 @@ namespace {
       return Addr;
     }
 
-    void reserveAllocationSpace(uintptr_t CodeSize, uint32_t CodeAlign,
-                                uintptr_t RODataSize, uint32_t RODataAlign,
+    void reserveAllocationSpace(uintptr_t CodeSize, Align CodeAlign,
+                                uintptr_t RODataSize, Align RODataAlign,
                                 uintptr_t RWDataSize,
-                                uint32_t RWDataAlign) override {
+                                Align RWDataAlign) override {
       m_Code.setAllocation(
-          Super::allocateCodeSection(CodeSize, CodeAlign,
+          Super::allocateCodeSection(CodeSize, CodeAlign.value(),
                                      /*SectionID=*/0,
                                      /*SectionName=*/"codeReserve"),
           CodeSize);
       m_ROData.setAllocation(
-          Super::allocateDataSection(RODataSize, RODataAlign,
+          Super::allocateDataSection(RODataSize, RODataAlign.value(),
                                      /*SectionID=*/0,
                                      /*SectionName=*/"rodataReserve",
                                      /*IsReadOnly=*/true),
           RODataSize);
       m_RWData.setAllocation(
-          Super::allocateDataSection(RWDataSize, RWDataAlign,
+          Super::allocateDataSection(RWDataSize, RWDataAlign.value(),
                                      /*SectionID=*/0,
                                      /*SectionName=*/"rwataReserve",
                                      /*IsReadOnly=*/false),
@@ -196,114 +194,24 @@ namespace {
   };
 
   /// A JITLinkMemoryManager for Cling that never frees its allocations.
-  class ClingJITLinkMemoryManager : public JITLinkMemoryManager {
+  class ClingJITLinkMemoryManager : public InProcessMemoryManager {
   public:
-    Expected<std::unique_ptr<Allocation>>
-    allocate(const JITLinkDylib* JD,
-             const SegmentsRequestMap& Request) override {
-      // A copy of InProcessMemoryManager::allocate with an empty implementation
-      // of IPMMAlloc::deallocate.
+    using InProcessMemoryManager::InProcessMemoryManager;
 
-      using AllocationMap = DenseMap<unsigned, sys::MemoryBlock>;
+    void deallocate(std::vector<FinalizedAlloc> Allocs,
+                    OnDeallocatedFunction OnDeallocated) override {
+      // Disabled until CallFunc is informed about unloading, and can
+      // re-generate the wrapper (if the decl is still available). See
+      // https://github.com/root-project/root/issues/10898
 
-      // Local class for allocation.
-      class IPMMAlloc : public Allocation {
-      public:
-        IPMMAlloc(AllocationMap SegBlocks) : SegBlocks(std::move(SegBlocks)) {}
-        MutableArrayRef<char> getWorkingMemory(ProtectionFlags Seg) override {
-          assert(SegBlocks.count(Seg) && "No allocation for segment");
-          return {static_cast<char*>(SegBlocks[Seg].base()),
-                  SegBlocks[Seg].allocatedSize()};
-        }
-        JITTargetAddress getTargetMemory(ProtectionFlags Seg) override {
-          assert(SegBlocks.count(Seg) && "No allocation for segment");
-          return pointerToJITTargetAddress(SegBlocks[Seg].base());
-        }
-        void finalizeAsync(FinalizeContinuation OnFinalize) override {
-          OnFinalize(applyProtections());
-        }
-        Error deallocate() override {
-          // Disabled until CallFunc is informed about unloading, and can
-          // re-generate the wrapper (if the decl is still available). See
-          // https://github.com/root-project/root/issues/10898
-          return Error::success();
-        }
-
-      private:
-        Error applyProtections() {
-          for (auto& KV : SegBlocks) {
-            auto& Prot = KV.first;
-            auto& Block = KV.second;
-            if (auto EC = sys::Memory::protectMappedMemory(Block, Prot))
-              return errorCodeToError(EC);
-            if (Prot & sys::Memory::MF_EXEC)
-              sys::Memory::InvalidateInstructionCache(Block.base(),
-                                                      Block.allocatedSize());
-          }
-          return Error::success();
-        }
-
-        AllocationMap SegBlocks;
-      };
-
-      if (!isPowerOf2_64((uint64_t)sys::Process::getPageSizeEstimate()))
-        return make_error<StringError>("Page size is not a power of 2",
-                                       inconvertibleErrorCode());
-
-      AllocationMap Blocks;
-      const sys::Memory::ProtectionFlags ReadWrite =
-          static_cast<sys::Memory::ProtectionFlags>(sys::Memory::MF_READ |
-                                                    sys::Memory::MF_WRITE);
-
-      // Compute the total number of pages to allocate.
-      size_t TotalSize = 0;
-      for (auto& KV : Request) {
-        const auto& Seg = KV.second;
-
-        if (Seg.getAlignment() > sys::Process::getPageSizeEstimate())
-          return make_error<StringError>("Cannot request higher than page "
-                                         "alignment",
-                                         inconvertibleErrorCode());
-
-        TotalSize = alignTo(TotalSize, sys::Process::getPageSizeEstimate());
-        TotalSize += Seg.getContentSize();
-        TotalSize += Seg.getZeroFillSize();
+      // We still have to release the allocations which resets their addresses
+      // to FinalizedAlloc::InvalidAddr, or the assertion in ~FinalizedAlloc
+      // will be unhappy...
+      for (auto &Alloc : Allocs) {
+        Alloc.release();
       }
-
-      // Allocate one slab to cover all the segments.
-      std::error_code EC;
-      auto SlabRemaining =
-          sys::Memory::allocateMappedMemory(TotalSize, nullptr, ReadWrite, EC);
-
-      if (EC)
-        return errorCodeToError(EC);
-
-      // Allocate segment memory from the slab.
-      for (auto& KV : Request) {
-
-        const auto& Seg = KV.second;
-
-        uint64_t SegmentSize =
-            alignTo(Seg.getContentSize() + Seg.getZeroFillSize(),
-                    sys::Process::getPageSizeEstimate());
-        assert(SlabRemaining.allocatedSize() >= SegmentSize &&
-               "Mapping exceeds allocation");
-
-        sys::MemoryBlock SegMem(SlabRemaining.base(), SegmentSize);
-        SlabRemaining =
-            sys::MemoryBlock((char*)SlabRemaining.base() + SegmentSize,
-                             SlabRemaining.allocatedSize() - SegmentSize);
-
-        // Zero out the zero-fill memory.
-        memset(static_cast<char*>(SegMem.base()) + Seg.getContentSize(), 0,
-               Seg.getZeroFillSize());
-
-        // Record the block for this segment.
-        Blocks[KV.first] = std::move(SegMem);
-      }
-
-      return std::unique_ptr<InProcessMemoryManager::Allocation>(
-          new IPMMAlloc(std::move(Blocks)));
+      // Pretend we successfully deallocated everything...
+      OnDeallocated(Error::success());
     }
   };
 
@@ -416,7 +324,7 @@ Error RTDynamicLibrarySearchGenerator::tryToGenerate(
 /// is called, which in turn may be used to provide lookup across different
 /// IncrementalJIT instances.
 class DelegateGenerator : public DefinitionGenerator {
-  using LookupFunc = std::function<Expected<JITEvaluatedSymbol>(StringRef)>;
+  using LookupFunc = std::function<Expected<llvm::orc::ExecutorAddr>(StringRef)>;
   LookupFunc lookup;
 
 public:
@@ -430,7 +338,9 @@ public:
       auto Addr = lookup(*KV.first);
       if (auto Err = Addr.takeError())
         return Err;
-      Symbols[KV.first] = Addr.get();
+      Symbols[KV.first] = JITEvaluatedSymbol(
+          Addr->getValue(),
+          JITSymbolFlags::Exported);
     }
     if (Symbols.empty())
       return Error::success();
@@ -444,7 +354,8 @@ static bool UseJITLink(const Triple& TT) {
   // LLJITBuilderState::prepareForConstruction.
   if (TT.getArch() == Triple::riscv64 ||
       (TT.isOSBinFormatMachO() &&
-       (TT.getArch() == Triple::aarch64 || TT.getArch() == Triple::x86_64))) {
+       (TT.getArch() == Triple::aarch64 || TT.getArch() == Triple::x86_64)) ||
+      (TT.isOSBinFormatELF() && TT.getArch() == Triple::ppc64le)) {
     jitLink = true;
   }
   // Finally, honor the user's choice by setting an environment variable.
@@ -553,16 +464,17 @@ IncrementalJIT::IncrementalJIT(
       // memory segments; the default InProcessMemoryManager (which is mostly
       // copied above) already does slab allocation to keep all segments
       // together which is needed for exception handling support.
+      unsigned PageSize = *sys::Process::getPageSize();
       auto ObjLinkingLayer = std::make_unique<ObjectLinkingLayer>(
-          ES, std::make_unique<ClingJITLinkMemoryManager>());
+          ES, std::make_unique<ClingJITLinkMemoryManager>(PageSize));
       ObjLinkingLayer->addPlugin(std::make_unique<EHFrameRegistrationPlugin>(
           ES, std::make_unique<InProcessEHFrameRegistrar>()));
       return ObjLinkingLayer;
     }
 
-    m_MMapper = std::make_unique<ClingMMapper>();
-    auto GetMemMgr = [this]() {
-      return std::make_unique<ClingMemoryManager>(*m_MMapper);
+    auto MMapper = std::make_unique<ClingMMapper>();
+    auto GetMemMgr = [MMapper = std::move(MMapper)]() {
+      return std::make_unique<ClingMemoryManager>(*MMapper);
     };
     auto Layer =
         std::make_unique<RTDyldObjectLinkingLayer>(ES, std::move(GetMemMgr));
@@ -673,11 +585,9 @@ IncrementalJIT::IncrementalJIT(
   Jit->getExecutionSession().setErrorReporter(ErrorReporter);
 }
 
-IncrementalJIT::~IncrementalJIT() = default;
-
 std::unique_ptr<llvm::orc::DefinitionGenerator> IncrementalJIT::getGenerator() {
   return std::make_unique<DelegateGenerator>(
-      [&](StringRef UnmangledName) { return Jit->lookup(UnmangledName); });
+      [&](StringRef Name) { return Jit->lookupLinkerMangled(Name); });
 }
 
 void IncrementalJIT::addModule(Transaction& T) {
@@ -777,11 +687,11 @@ void* IncrementalJIT::getSymbolAddress(StringRef Name, bool IncludeHostSymbols){
   if (!IncludeHostSymbols)
     G.lock();
 
-  std::pair<llvm::StringMapIterator<llvm::NoneType>, bool> insertInfo;
+  std::pair<llvm::StringMapIterator<std::nullopt_t>, bool> insertInfo;
   if (!IncludeHostSymbols)
     insertInfo = m_ForbidDlSymbols.insert(Name);
 
-  Expected<JITEvaluatedSymbol> Symbol = Jit->lookup(Name);
+  Expected<llvm::orc::ExecutorAddr> Symbol = Jit->lookup(Name);
 
   // If m_ForbidDlSymbols already contained Name before we tried to insert it
   // then some calling frame has added it and will remove it later because its
@@ -796,7 +706,7 @@ void* IncrementalJIT::getSymbolAddress(StringRef Name, bool IncludeHostSymbols){
     return nullptr;
   }
 
-  return jitTargetAddressToPointer<void*>(Symbol->getAddress());
+  return jitTargetAddressToPointer<void*>(Symbol->getValue());
 }
 
 bool IncrementalJIT::doesSymbolAlreadyExist(StringRef UnmangledName) {
