@@ -69,11 +69,14 @@ using namespace clang;
 
 namespace {
 
-  // Forward cxa_atexit for global d'tors.
-  static int local_cxa_atexit(void (*func) (void*), void* arg,
-                              cling::Interpreter* Interp) {
-    Interp->AddAtExitFunc(func, arg);
-    return 0;
+  static void registerCxaAtExitHelper(void *Self, void (*F)(void *), void *Ctx,
+                                      void *DSOHandle) {
+    static_cast<cling::Interpreter *>(Self)->AddAtExitFunc(F, Ctx);
+  }
+
+  static void registerAtExitHelper(void *Self, void *DSOHandle, void (*F)()) {
+    static_cast<cling::Interpreter *>(Self)->AddAtExitFunc(
+        reinterpret_cast<void (*)(void *)>(F), nullptr);
   }
 
   static cling::Interpreter::ExecutionResult
@@ -190,7 +193,7 @@ namespace cling {
 
     // Disable suggestions for ROOT
     bool showSuggestions =
-        !llvm::StringRef(ClingStringify(CLING_VERSION)).startswith("ROOT");
+        !llvm::StringRef(ClingStringify(CLING_VERSION)).starts_with("ROOT");
 
     std::unique_ptr<InterpreterCallbacks> AutoLoadCB(
         new AutoloadCallback(&Interp, showSuggestions));
@@ -257,7 +260,7 @@ namespace cling {
     if (!m_LookupHelper)
       return;
 
-    if (!isInSyntaxOnlyMode()) {
+    if (!isInSyntaxOnlyMode() && !m_Opts.CompilerOpts.CUDADevice) {
       m_Executor.reset(new IncrementalExecutor(SemaRef.Diags, *getCI(),
         extraLibHandle, m_Opts.Verbose()));
 
@@ -312,7 +315,8 @@ namespace cling {
     Initialize(noRuntime || m_Opts.NoRuntime, isInSyntaxOnlyMode(), Syms);
 
     // Commit the transactions, now that gCling is set up. It is needed for
-    // static initialization in these transactions through local_cxa_atexit().
+    // static initialization in these transactions through
+    // registerCxaAtExitHelper().
     for (auto&& I: IncrParserTransactions)
       m_IncrParser->commitTransaction(I);
 
@@ -400,7 +404,7 @@ namespace cling {
     // and the ASTContext. For that to happen we shut down the IncrementalParser
     // explicitly, before the implicit destruction (through the unique_ptr) of
     // the callbacks.
-    m_IncrParser.reset(0);
+    m_IncrParser.reset(nullptr);
   }
 
   Transaction* Interpreter::Initialize(bool NoRuntime, bool SyntaxOnly,
@@ -491,13 +495,7 @@ namespace cling {
     if (EmitDefinitions)
       Strm << "#define __dso_handle ((void*)" << ThisP << ")\n";
 
-    // C atexit, std::atexit
-    Strm << Linkage << " int atexit(void(*f)()) " << Attr;
-    if (EmitDefinitions)
-      Strm << " { return __cxa_atexit((void(*)(void*))f, 0, __dso_handle); }\n";
-    else
-      Strm << ";\n";
-    Globals.push_back("atexit");
+    // C atexit, std::atexit -- registered by GenericLLVMIRPlatformSupport
 
     // C++ 11 at_quick_exit, std::at_quick_exit
     if (LangOpts.CPlusPlus && LangOpts.CPlusPlus11) {
@@ -539,10 +537,24 @@ namespace cling {
 #endif
 
     if (!SyntaxOnly) {
-      // Override the native symbols now, before anything can be emitted.
-      m_Executor->replaceSymbol("__cxa_atexit",
-                            utils::FunctionToVoidPtr(&local_cxa_atexit));
-      // __dso_handle is inserted for the link phase, as macro is useless then
+      // Override the helper symbols injected by GenericLLVMIRPlatformSupport,
+      // before anything can be emitted.
+      m_Executor->replaceSymbol("__lljit.cxa_atexit_helper",
+                            utils::FunctionToVoidPtr(&registerCxaAtExitHelper));
+      m_Executor->replaceSymbol("__lljit.atexit_helper",
+                            utils::FunctionToVoidPtr(&registerAtExitHelper));
+      // Replace __lljit.platform_support_instance which is passed as Self
+      // to the helper functions. We cannot easily replace __dso_handle (in
+      // the code already generated by GenericLLVMIRPlatformSupport) because
+      // it is wired up as a constant (for the current JITDylib)...
+      m_Executor->replaceSymbol("__lljit.platform_support_instance", this);
+      // Still insert __dso_handle for the link phase, as macro is useless then.
+      // TODO: Is this still relevant now that registerCxaAtExitHelper uses the
+      // Self parameter and the DSOHandle argument is basically unused? The only
+      // weird procedure I found to make this matter is:
+      //     #undef __dso_handle
+      //     extern "C" void *__dso_handle;
+      // then cling::runtime::gCling is identical to &__dso_handle.
       m_Executor->replaceSymbol("__dso_handle", this);
 
 #ifdef _MSC_VER
@@ -872,7 +884,8 @@ namespace cling {
     Preprocessor& PP = getCI()->getPreprocessor();
     HeaderSearch &HS = PP.getHeaderSearchInfo();
 
-    if (Module *M = HS.lookupModule(moduleName, /*AllowSearch*/true,
+    if (Module *M = HS.lookupModule(moduleName, SourceLocation(),
+                                    /*AllowSearch*/true,
                                     /*AllowExtraSearch*/ true))
       return loadModule(M, complain);
 
@@ -1112,7 +1125,7 @@ namespace cling {
   }
 
   bool Interpreter::isUniqueName(llvm::StringRef name) {
-    return name.startswith(utils::Synthesize::UniquePrefix);
+    return name.starts_with(utils::Synthesize::UniquePrefix);
   }
 
   clang::SourceLocation Interpreter::getSourceLocation(bool skipWrapper) const {
@@ -1268,7 +1281,7 @@ namespace cling {
                      clang::diag::Severity::Warning, SourceLocation());
 
     if (CR != cling::Interpreter::kSuccess)
-      return 0;
+      return nullptr;
 
     for (cling::Transaction::const_iterator I = T->decls_begin(),
            E = T->decls_end(); I != E; ++I) {
@@ -1286,7 +1299,7 @@ namespace cling {
         }
       }
     }
-    return 0;
+    return nullptr;
   }
 
   void*
@@ -1297,7 +1310,7 @@ namespace cling {
     //
 
     if (isInSyntaxOnlyMode())
-      return 0;
+      return nullptr;
 
     if (ifUnique) {
       if (void* Addr = (void*)getAddressOfGlobal(name)) {
@@ -1308,7 +1321,7 @@ namespace cling {
     Transaction* T = nullptr;
     const FunctionDecl* FD = DeclareCFunction(name, code, withAccessControl, T);
     if (!FD || !T)
-      return 0;
+      return nullptr;
 
     //  Get the wrapper function pointer from the ExecutionEngine (the JIT).
     return m_Executor->getPointerToGlobalFromJIT(name);
@@ -1441,16 +1454,16 @@ namespace cling {
     // Clang doc says:
     // "LookupFrom is set when this is a \#include_next directive, it
     // specifies the file to start searching from."
-    const DirectoryLookup* FromDir = 0;
-    const FileEntry* FromFile = 0;
-    const DirectoryLookup* CurDir = 0;
+    ConstSearchDirIterator FromDir = nullptr;
+    const FileEntry* FromFile = nullptr;
+    ConstSearchDirIterator* CurDir = nullptr;
     Preprocessor& PP = getCI()->getPreprocessor();
     // PP::LookupFile uses it to issue 'nice' diagnostic
     SourceLocation fileNameLoc;
     auto FE = PP.LookupFile(fileNameLoc, canonicalFile, isAngled, FromDir,
-                            FromFile, CurDir, /*SearchPath*/0,
-                            /*RelativePath*/ 0, /*suggestedModule*/0,
-                            0 /*IsMapped*/, /*IsFrameworkFound*/ nullptr,
+                            FromFile, CurDir, /*SearchPath*/nullptr,
+                            /*RelativePath*/ nullptr, /*suggestedModule*/nullptr,
+                            /*IsMapped*/ nullptr, /*IsFrameworkFound*/ nullptr,
                             /*SkipCache*/ false, /*OpenFile*/ false,
                             /*CacheFail*/ false);
     if (FE)
@@ -1749,7 +1762,7 @@ namespace cling {
                                         bool* fromJIT /*=0*/) const {
     // Return a symbol's address, and whether it was jitted.
     if (isInSyntaxOnlyMode())
-      return 0;
+      return nullptr;
     return m_Executor->getAddressOfGlobal(SymName, fromJIT);
   }
 
@@ -1841,7 +1854,7 @@ namespace cling {
         }();
         if (!ret.isValid()) {
           std::string msg = "Error evaluating expression ";
-          CompilationException::throwingHandler(nullptr, msg + DEI->getExpr(),
+          CompilationException::throwingHandler(nullptr, (msg + DEI->getExpr()).c_str(),
                                                 false /*backtrace*/);
         }
         return ret;
