@@ -14,21 +14,26 @@
  *************************************************************************/
 
 #include <ROOT/RConfig.hxx>
+#include <ROOT/RError.hxx>
 
 #include "ROOT/RMiniFile.hxx"
 
 #include <ROOT/RRawFile.hxx>
-#include <ROOT/RNTuple.hxx> // For converting an anchor to an RNTuple object
 #include <ROOT/RNTupleZip.hxx>
 
 #include <TError.h>
 #include <TFile.h>
 #include <TKey.h>
 
+#include <xxhash.h>
+
 #include <algorithm>
+#include <cerrno>
 #include <cstdio>
 #include <cstring>
 #include <iostream>
+#include <memory>
+#include <new>
 #include <string>
 #include <utility>
 #include <chrono>
@@ -119,25 +124,27 @@ public:
 
 constexpr std::int32_t ChecksumRNTupleClass() {
    const char ident[] = "ROOT::Experimental::RNTuple"
-                        "fChecksum"
-                        "int"
-                        "fVersion"
-                        "unsigned int"
-                        "fSize"
-                        "unsigned int"
+                        "fVersionEpoch"
+                        "unsigned short"
+                        "fVersionMajor"
+                        "unsigned short"
+                        "fVersionMinor"
+                        "unsigned short"
+                        "fVersionPatch"
+                        "unsigned short"
                         "fSeekHeader"
                         "unsigned long"
                         "fNBytesHeader"
-                        "unsigned int"
+                        "unsigned long"
                         "fLenHeader"
-                        "unsigned int"
+                        "unsigned long"
                         "fSeekFooter"
                         "unsigned long"
                         "fNBytesFooter"
-                        "unsigned int"
+                        "unsigned long"
                         "fLenFooter"
-                        "unsigned int"
-                        "fReserved"
+                        "unsigned long"
+                        "fChecksum"
                         "unsigned long";
    std::int32_t id = 0;
    for (unsigned i = 0; i < (sizeof(ident) - 1); i++)
@@ -149,15 +156,23 @@ constexpr std::int32_t ChecksumRNTupleClass() {
 #pragma pack(push, 1)
 /// A name (type, identifies, ...) in the TFile binary format
 struct RTFString {
-   char fLName{0};
+   unsigned char fLName{0};
    char fData[255];
    RTFString() = default;
    RTFString(const std::string &str) {
-      R__ASSERT(str.length() < 256);
+      // The length of strings with 255 characters and longer are encoded with a 32-bit integer following the first
+      // byte. This is currently not handled.
+      R__ASSERT(str.length() < 255);
       fLName = str.length();
       memcpy(fData, str.data(), fLName);
    }
-   char GetSize() const { return 1 + fLName; }
+   std::size_t GetSize() const
+   {
+      // A length of 255 is special and means that the first byte is followed by a 32-bit integer with the actual
+      // length.
+      R__ASSERT(fLName != 255);
+      return 1 + fLName;
+   }
 };
 
 /// The timestamp format used in TFile; the default constructor initializes with the current time
@@ -197,8 +212,10 @@ struct RTFKey {
    RTFKey() : fInfoShort() {}
    RTFKey(std::uint64_t seekKey, std::uint64_t seekPdir,
           const RTFString &clName, const RTFString &objName, const RTFString &titleName,
-          std::uint32_t szObjInMem, std::uint32_t szObjOnDisk = 0)
+          std::size_t szObjInMem, std::size_t szObjOnDisk = 0)
    {
+      R__ASSERT(szObjInMem < std::numeric_limits<std::int32_t>::max());
+      R__ASSERT(szObjOnDisk < std::numeric_limits<std::int32_t>::max());
       fObjLen = szObjInMem;
       if ((seekKey > static_cast<unsigned int>(std::numeric_limits<std::int32_t>::max())) ||
           (seekPdir > static_cast<unsigned int>(std::numeric_limits<std::int32_t>::max())))
@@ -226,6 +243,7 @@ struct RTFKey {
       fInfoLong.fSeekKey = seekKey;
       fInfoLong.fSeekPdir = seekPdir;
       fKeyHeaderSize = fKeyHeaderSize + sizeof(fInfoLong) - sizeof(fInfoShort);
+      fKeyLen = fKeyLen + sizeof(fInfoLong) - sizeof(fInfoShort);
       fNbytes = fNbytes + sizeof(fInfoLong) - sizeof(fInfoShort);
       fVersion = fVersion + 1000;
    }
@@ -433,67 +451,88 @@ struct RTFObject {
    explicit RTFObject(std::uint32_t bits) : fBits(bits) {}
 };
 
-/// Streamer info for data member RNTuple::fChecksum
-struct RTFStreamerElementChecksum {
-   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerElementChecksum) - sizeof(RUInt32BE))};
+/// Streamer info for data member RNTuple::fVersionEpoch
+struct RTFStreamerElementVersionEpoch {
+   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerElementVersionEpoch) - sizeof(RUInt32BE))};
    RUInt16BE fVersion{4};
 
-   RUInt32BE fByteCountNamed{0x40000000 | (sizeof(RUInt16BE) + sizeof(RTFObject) + 11)};
+   RUInt32BE fByteCountNamed{0x40000000 | (sizeof(RUInt16BE) + sizeof(RTFObject) + 15)};
    RUInt16BE fVersionNamed{1};
    RTFObject fObjectNamed{0x02000000 | 0x01000000};
-   char fLName = 9;
-   char fName[9]{'f', 'C', 'h', 'e', 'c', 'k', 's', 'u', 'm'};
+   char fLName = 13;
+   char fName[13]{'f', 'V', 'e', 'r', 's', 'i', 'o', 'n', 'E', 'p', 'o', 'c', 'h'};
    char fLTitle = 0;
 
-   RUInt32BE fType{3};
-   RUInt32BE fSize{4};
+   RUInt32BE fType{12};
+   RUInt32BE fSize{2};
    RUInt32BE fArrLength{0};
    RUInt32BE fArrDim{0};
    char fMaxIndex[20]{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
-   char fLTypeName = 3;
-   char fTypeName[3]{'i', 'n', 't'};
+   char fLTypeName = 14;
+   char fTypeName[14]{'u', 'n', 's', 'i', 'g', 'n', 'e', 'd', ' ', 's', 'h', 'o', 'r', 't'};
 };
 
-/// Streamer info for data member RNTuple::fVersion
-struct RTFStreamerElementVersion {
-   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerElementVersion) - sizeof(RUInt32BE))};
+/// Streamer info for data member RNTuple::fVersionMajor
+struct RTFStreamerElementVersionMajor {
+   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerElementVersionMajor) - sizeof(RUInt32BE))};
    RUInt16BE fVersion{4};
 
-   RUInt32BE fByteCountNamed{0x40000000 | (sizeof(RUInt16BE) + sizeof(RTFObject) + 10)};
+   RUInt32BE fByteCountNamed{0x40000000 | (sizeof(RUInt16BE) + sizeof(RTFObject) + 15)};
    RUInt16BE fVersionNamed{1};
    RTFObject fObjectNamed{0x02000000 | 0x01000000};
-   char fLName = 8;
-   char fName[8]{ 'f', 'V', 'e', 'r', 's', 'i', 'o', 'n' };
+   char fLName = 13;
+   char fName[13]{'f', 'V', 'e', 'r', 's', 'i', 'o', 'n', 'M', 'a', 'j', 'o', 'r'};
    char fLTitle = 0;
 
-   RUInt32BE fType{13};
-   RUInt32BE fSize{4};
+   RUInt32BE fType{12};
+   RUInt32BE fSize{2};
    RUInt32BE fArrLength{0};
    RUInt32BE fArrDim{0};
    char fMaxIndex[20]{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
-   char fLTypeName = 12;
-   char fTypeName[12]{ 'u', 'n', 's', 'i', 'g', 'n', 'e', 'd', ' ', 'i', 'n', 't' };
+   char fLTypeName = 14;
+   char fTypeName[14]{'u', 'n', 's', 'i', 'g', 'n', 'e', 'd', ' ', 's', 'h', 'o', 'r', 't'};
 };
 
-/// Streamer info for data member RNTuple::fSize
-struct RTFStreamerElementSize {
-   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerElementSize) - sizeof(RUInt32BE))};
+/// Streamer info for data member RNTuple::fVersionMajor
+struct RTFStreamerElementVersionMinor {
+   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerElementVersionMinor) - sizeof(RUInt32BE))};
    RUInt16BE fVersion{4};
 
-   RUInt32BE fByteCountNamed{0x40000000 | (sizeof(RUInt16BE) + sizeof(RTFObject) + 7)};
+   RUInt32BE fByteCountNamed{0x40000000 | (sizeof(RUInt16BE) + sizeof(RTFObject) + 15)};
    RUInt16BE fVersionNamed{1};
    RTFObject fObjectNamed{0x02000000 | 0x01000000};
-   char fLName = 5;
-   char fName[5]{ 'f', 'S', 'i', 'z', 'e' };
+   char fLName = 13;
+   char fName[13]{'f', 'V', 'e', 'r', 's', 'i', 'o', 'n', 'M', 'i', 'n', 'o', 'r'};
    char fLTitle = 0;
 
-   RUInt32BE fType{13};
-   RUInt32BE fSize{4};
+   RUInt32BE fType{12};
+   RUInt32BE fSize{2};
    RUInt32BE fArrLength{0};
    RUInt32BE fArrDim{0};
    char fMaxIndex[20]{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
-   char fLTypeName = 12;
-   char fTypeName[12]{ 'u', 'n', 's', 'i', 'g', 'n', 'e', 'd', ' ', 'i', 'n', 't' };
+   char fLTypeName = 14;
+   char fTypeName[14]{'u', 'n', 's', 'i', 'g', 'n', 'e', 'd', ' ', 's', 'h', 'o', 'r', 't'};
+};
+
+/// Streamer info for data member RNTuple::fVersionPatch
+struct RTFStreamerElementVersionPatch {
+   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerElementVersionPatch) - sizeof(RUInt32BE))};
+   RUInt16BE fVersion{4};
+
+   RUInt32BE fByteCountNamed{0x40000000 | (sizeof(RUInt16BE) + sizeof(RTFObject) + 15)};
+   RUInt16BE fVersionNamed{1};
+   RTFObject fObjectNamed{0x02000000 | 0x01000000};
+   char fLName = 13;
+   char fName[13]{'f', 'V', 'e', 'r', 's', 'i', 'o', 'n', 'P', 'a', 't', 'c', 'h'};
+   char fLTitle = 0;
+
+   RUInt32BE fType{12};
+   RUInt32BE fSize{2};
+   RUInt32BE fArrLength{0};
+   RUInt32BE fArrDim{0};
+   char fMaxIndex[20]{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
+   char fLTypeName = 14;
+   char fTypeName[14]{'u', 'n', 's', 'i', 'g', 'n', 'e', 'd', ' ', 's', 'h', 'o', 'r', 't'};
 };
 
 /// Streamer info for data member RNTuple::fSeekHeader
@@ -531,13 +570,13 @@ struct RTFStreamerElementNBytesHeader {
    char fName[13]{ 'f', 'N', 'B', 'y', 't', 'e', 's', 'H', 'e', 'a', 'd', 'e', 'r' };
    char fLTitle = 0;
 
-   RUInt32BE fType{13};
-   RUInt32BE fSize{4};
+   RUInt32BE fType{14};
+   RUInt32BE fSize{8};
    RUInt32BE fArrLength{0};
    RUInt32BE fArrDim{0};
    char fMaxIndex[20]{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
-   char fLTypeName = 12;
-   char fTypeName[12]{ 'u', 'n', 's', 'i', 'g', 'n', 'e', 'd', ' ', 'i', 'n', 't' };
+   char fLTypeName = 13;
+   char fTypeName[13]{'u', 'n', 's', 'i', 'g', 'n', 'e', 'd', ' ', 'l', 'o', 'n', 'g'};
 };
 
 /// Streamer info for data member RNTuple::fLenHeader
@@ -553,13 +592,13 @@ struct RTFStreamerElementLenHeader {
    char fName[10]{ 'f', 'L', 'e', 'n', 'H', 'e', 'a', 'd', 'e', 'r' };
    char fLTitle = 0;
 
-   RUInt32BE fType{13};
-   RUInt32BE fSize{4};
+   RUInt32BE fType{14};
+   RUInt32BE fSize{8};
    RUInt32BE fArrLength{0};
    RUInt32BE fArrDim{0};
    char fMaxIndex[20]{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
-   char fLTypeName = 12;
-   char fTypeName[12]{ 'u', 'n', 's', 'i', 'g', 'n', 'e', 'd', ' ', 'i', 'n', 't' };
+   char fLTypeName = 13;
+   char fTypeName[13]{'u', 'n', 's', 'i', 'g', 'n', 'e', 'd', ' ', 'l', 'o', 'n', 'g'};
 };
 
 /// Streamer info for data member RNTuple::fSeekFooter
@@ -596,13 +635,13 @@ struct RTFStreamerElementNBytesFooter {
    char fName[13]{ 'f', 'N', 'B', 'y', 't', 'e', 's', 'F', 'o', 'o', 't', 'e', 'r' };
    char fLTitle = 0;
 
-   RUInt32BE fType{13};
-   RUInt32BE fSize{4};
+   RUInt32BE fType{14};
+   RUInt32BE fSize{8};
    RUInt32BE fArrLength{0};
    RUInt32BE fArrDim{0};
    char fMaxIndex[20]{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
-   char fLTypeName = 12;
-   char fTypeName[12]{ 'u', 'n', 's', 'i', 'g', 'n', 'e', 'd', ' ', 'i', 'n', 't' };
+   char fLTypeName = 13;
+   char fTypeName[13]{'u', 'n', 's', 'i', 'g', 'n', 'e', 'd', ' ', 'l', 'o', 'n', 'g'};
 };
 
 /// Streamer info for data member RNTuple::fLenFooter
@@ -617,18 +656,18 @@ struct RTFStreamerElementLenFooter {
    char fName[10]{ 'f', 'L', 'e', 'n', 'F', 'o', 'o', 't', 'e', 'r' };
    char fLTitle = 0;
 
-   RUInt32BE fType{13};
-   RUInt32BE fSize{4};
+   RUInt32BE fType{14};
+   RUInt32BE fSize{8};
    RUInt32BE fArrLength{0};
    RUInt32BE fArrDim{0};
    char fMaxIndex[20]{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
-   char fLTypeName = 12;
-   char fTypeName[12]{ 'u', 'n', 's', 'i', 'g', 'n', 'e', 'd', ' ', 'i', 'n', 't' };
+   char fLTypeName = 13;
+   char fTypeName[13]{'u', 'n', 's', 'i', 'g', 'n', 'e', 'd', ' ', 'l', 'o', 'n', 'g'};
 };
 
-/// Streamer info for data member RNTuple::fReserved
-struct RTFStreamerElementReserved {
-   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerElementReserved) - sizeof(RUInt32BE))};
+/// Streamer info for data member RNTuple::fChecksum
+struct RTFStreamerElementChecksum {
+   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerElementChecksum) - sizeof(RUInt32BE))};
    RUInt16BE fVersion{4};
 
    RUInt32BE fByteCountNamed{0x40000000 |
@@ -636,7 +675,7 @@ struct RTFStreamerElementReserved {
    RUInt16BE fVersionNamed{1};
    RTFObject fObjectNamed{0x02000000 | 0x01000000};
    char fLName = 9;
-   char fName[9]{ 'f', 'R', 'e', 's', 'e', 'r', 'v', 'e', 'd' };
+   char fName[9]{'f', 'C', 'h', 'e', 'c', 'k', 's', 'u', 'm'};
    char fLTitle = 0;
 
    RUInt32BE fType{14};
@@ -648,33 +687,42 @@ struct RTFStreamerElementReserved {
    char fTypeName[13]{ 'u', 'n', 's', 'i', 'g', 'n', 'e', 'd', ' ', 'l', 'o', 'n', 'g' };
 };
 
-/// Streamer info frame for data member RNTuple::fChecksum
-struct RTFStreamerChecksum {
-   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerChecksum) - sizeof(RUInt32BE))};
+/// Streamer info frame for data member RNTuple::fVersionEpoch
+struct RTFStreamerVersionEpoch {
+   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerVersionEpoch) - sizeof(RUInt32BE))};
    RUInt32BE fNewClassTag{0xffffffff};
    char fClassName[19]{'T', 'S', 't', 'r', 'e', 'a', 'm', 'e', 'r', 'B', 'a', 's', 'i', 'c', 'T', 'y', 'p', 'e', '\0'};
-   RUInt32BE fByteCountRemaining{0x40000000 | (sizeof(RTFStreamerChecksum) - 2 * sizeof(RUInt32BE) -
+   RUInt32BE fByteCountRemaining{0x40000000 | (sizeof(RTFStreamerVersionEpoch) - 2 * sizeof(RUInt32BE) -
                                                19 /* strlen(fClassName) + 1 */ - sizeof(RUInt32BE))};
    RUInt16BE fVersion{2};
-   RTFStreamerElementChecksum fStreamerElementChecksum;
+   RTFStreamerElementVersionEpoch fStreamerElementVersionEpoch;
 };
 
-/// Streamer info frame for data member RNTuple::fVersion
-struct RTFStreamerVersion {
-   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerVersion) - sizeof(RUInt32BE))};
+/// Streamer info frame for data member RNTuple::fVersionMajor
+struct RTFStreamerVersionMajor {
+   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerVersionMajor) - sizeof(RUInt32BE))};
    RUInt32BE fClassTag{0x80000000}; // Fix-up after construction, or'd with 0x80000000
-   RUInt32BE fByteCountRemaining{0x40000000 | (sizeof(RTFStreamerVersion) - 3 * sizeof(RUInt32BE))};
+   RUInt32BE fByteCountRemaining{0x40000000 | (sizeof(RTFStreamerVersionMajor) - 3 * sizeof(RUInt32BE))};
    RUInt16BE fVersion{2};
-   RTFStreamerElementVersion fStreamerElementVersion;
+   RTFStreamerElementVersionMajor fStreamerElementVersionMajor;
 };
 
-/// Streamer info frame for data member RNTuple::fSize
-struct RTFStreamerSize {
-   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerSize) - sizeof(RUInt32BE))};
-   RUInt32BE fClassTag{0x80000000};  // Fix-up after construction, or'd with 0x80000000
-   RUInt32BE fByteCountRemaining{0x40000000 | (sizeof(RTFStreamerSize) - 3 * sizeof(RUInt32BE))};
+/// Streamer info frame for data member RNTuple::fVersionMinor
+struct RTFStreamerVersionMinor {
+   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerVersionMinor) - sizeof(RUInt32BE))};
+   RUInt32BE fClassTag{0x80000000}; // Fix-up after construction, or'd with 0x80000000
+   RUInt32BE fByteCountRemaining{0x40000000 | (sizeof(RTFStreamerVersionMinor) - 3 * sizeof(RUInt32BE))};
    RUInt16BE fVersion{2};
-   RTFStreamerElementSize fStreamerElementSize;
+   RTFStreamerElementVersionMinor fStreamerElementVersionMinor;
+};
+
+/// Streamer info frame for data member RNTuple::fVersionPatch
+struct RTFStreamerVersionPatch {
+   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerVersionPatch) - sizeof(RUInt32BE))};
+   RUInt32BE fClassTag{0x80000000};  // Fix-up after construction, or'd with 0x80000000
+   RUInt32BE fByteCountRemaining{0x40000000 | (sizeof(RTFStreamerVersionPatch) - 3 * sizeof(RUInt32BE))};
+   RUInt16BE fVersion{2};
+   RTFStreamerElementVersionPatch fStreamerElementVersionPatch;
 };
 
 /// Streamer info frame for data member RNTuple::fSeekHeader
@@ -731,13 +779,13 @@ struct RTFStreamerLenFooter {
    RTFStreamerElementLenFooter fStreamerElementLenFooter;
 };
 
-/// Streamer info frame for data member RNTuple::fReserved
-struct RTFStreamerReserved {
-   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerReserved) - sizeof(RUInt32BE))};
+/// Streamer info frame for data member RNTuple::fChecksum
+struct RTFStreamerChecksum {
+   RUInt32BE fByteCount{0x40000000 | (sizeof(RTFStreamerChecksum) - sizeof(RUInt32BE))};
    RUInt32BE fClassTag{0x80000000};  // Fix-up after construction, or'd with 0x80000000
-   RUInt32BE fByteCountRemaining{0x40000000 | (sizeof(RTFStreamerReserved) - 3 * sizeof(RUInt32BE))};
+   RUInt32BE fByteCountRemaining{0x40000000 | (sizeof(RTFStreamerChecksum) - 3 * sizeof(RUInt32BE))};
    RUInt16BE fVersion{2};
-   RTFStreamerElementReserved fStreamerElementReserved;
+   RTFStreamerElementChecksum fStreamerElementChecksum;
 };
 
 /// Streamer info for class RNTuple
@@ -760,7 +808,7 @@ struct RTFStreamerInfoObject {
    char fLTitle = 0;
 
    RInt32BE fChecksum{ChecksumRNTupleClass()};
-   RUInt32BE fVersionRNTuple{1};
+   RUInt32BE fVersionRNTuple{4};
 
    RUInt32BE fByteCountObjArr{0x40000000 |
       (sizeof(RUInt32BE) + 10 /* strlen(TObjArray) + 1 */ + sizeof(RUInt32BE) +
@@ -775,20 +823,21 @@ struct RTFStreamerInfoObject {
    RTFObject fObjectObjArr{0x02000000};
    char fNameObjArr{0};
 
-   RUInt32BE fNObjects{10};
+   RUInt32BE fNObjects{11};
    RUInt32BE fLowerBound{0};
 
    struct {
-      RTFStreamerChecksum fStreamerChecksum;
-      RTFStreamerVersion fStreamerVersion;
-      RTFStreamerSize fStreamerSize;
+      RTFStreamerVersionEpoch fStreamerVersionEpoch;
+      RTFStreamerVersionMajor fStreamerVersionMajor;
+      RTFStreamerVersionMinor fStreamerVersionMinor;
+      RTFStreamerVersionPatch fStreamerVersionPatch;
       RTFStreamerSeekHeader fStreamerSeekHeader;
       RTFStreamerNBytesHeader fStreamerNBytesHeader;
       RTFStreamerLenHeader fStreamerLenHeader;
       RTFStreamerSeekFooter fStreamerSeekFooter;
       RTFStreamerNBytesFooter fStreamerNBytesFooter;
       RTFStreamerLenFooter fStreamerLenFooter;
-      RTFStreamerReserved fStreamerReserved;
+      RTFStreamerChecksum fStreamerChecksum;
    } fStreamers;
 };
 
@@ -877,47 +926,39 @@ struct RTFUUID {
 /// A streamed RNTuple class
 struct RTFNTuple {
    RUInt32BE fByteCount{0x40000000 | (sizeof(RTFNTuple) - sizeof(fByteCount))};
-   RUInt16BE fVersionClass{3};
-   RInt32BE fChecksum{ChecksumRNTupleClass()};
-   RUInt32BE fVersionInternal{0};
-   RUInt32BE fSize{sizeof(ROOT::Experimental::Internal::RFileNTupleAnchor)};
+   RUInt16BE fVersionClass{4};
+   RUInt16BE fVersionEpoch{0};
+   RUInt16BE fVersionMajor{0};
+   RUInt16BE fVersionMinor{0};
+   RUInt16BE fVersionPatch{0};
    RUInt64BE fSeekHeader{0};
-   RUInt32BE fNBytesHeader{0};
-   RUInt32BE fLenHeader{0};
+   RUInt64BE fNBytesHeader{0};
+   RUInt64BE fLenHeader{0};
    RUInt64BE fSeekFooter{0};
-   RUInt32BE fNBytesFooter{0};
-   RUInt32BE fLenFooter{0};
-   RUInt64BE fReserved{0};
+   RUInt64BE fNBytesFooter{0};
+   RUInt64BE fLenFooter{0};
+   RUInt64BE fChecksum{0};
 
    RTFNTuple() = default;
-   explicit RTFNTuple(const ROOT::Experimental::Internal::RFileNTupleAnchor &inMemoryAnchor)
+   explicit RTFNTuple(const ROOT::Experimental::RNTuple &inMemoryAnchor)
    {
-      fVersionInternal = inMemoryAnchor.fVersion;
-      fSize = inMemoryAnchor.fSize;
-      fSeekHeader = inMemoryAnchor.fSeekHeader;
-      fNBytesHeader = inMemoryAnchor.fNBytesHeader;
-      fLenHeader = inMemoryAnchor.fLenHeader;
-      fSeekFooter = inMemoryAnchor.fSeekFooter;
-      fNBytesFooter = inMemoryAnchor.fNBytesFooter;
-      fLenFooter = inMemoryAnchor.fLenFooter;
-      fReserved = inMemoryAnchor.fReserved;
+      fVersionEpoch = inMemoryAnchor.GetVersionEpoch();
+      fVersionMajor = inMemoryAnchor.GetVersionMajor();
+      fVersionMinor = inMemoryAnchor.GetVersionMinor();
+      fVersionPatch = inMemoryAnchor.GetVersionPatch();
+      fSeekHeader = inMemoryAnchor.GetSeekHeader();
+      fNBytesHeader = inMemoryAnchor.GetNBytesHeader();
+      fLenHeader = inMemoryAnchor.GetLenHeader();
+      fSeekFooter = inMemoryAnchor.GetSeekFooter();
+      fNBytesFooter = inMemoryAnchor.GetNBytesFooter();
+      fLenFooter = inMemoryAnchor.GetLenFooter();
+      fChecksum = XXH3_64bits(GetPtrCkData(), GetSizeCkData());
    }
    std::uint32_t GetSize() const { return sizeof(RTFNTuple); }
-   ROOT::Experimental::Internal::RFileNTupleAnchor ToAnchor() const
-   {
-      ROOT::Experimental::Internal::RFileNTupleAnchor anchor;
-      anchor.fChecksum = fChecksum;
-      anchor.fVersion = fVersionInternal;
-      anchor.fSize = fSize;
-      anchor.fSeekHeader = fSeekHeader;
-      anchor.fNBytesHeader = fNBytesHeader;
-      anchor.fLenHeader = fLenHeader;
-      anchor.fSeekFooter = fSeekFooter;
-      anchor.fNBytesFooter = fNBytesFooter;
-      anchor.fLenFooter = fLenFooter;
-      anchor.fReserved = fReserved;
-      return anchor;
-   }
+   // The byte count and class version members are not checksummed
+   std::uint32_t GetOffsetCkData() { return sizeof(fByteCount) + sizeof(fVersionClass); }
+   std::uint32_t GetSizeCkData() { return GetSize() - GetOffsetCkData() - sizeof(fChecksum); }
+   unsigned char *GetPtrCkData() { return reinterpret_cast<unsigned char *>(this) + GetOffsetCkData(); }
 };
 
 /// The bare file global header
@@ -934,9 +975,9 @@ struct RBareFileHeader {
 #pragma pack(pop)
 
 /// The artifical class name shown for opaque RNTuple keys (see TBasket)
-static constexpr char const *kBlobClassName = "RBlob";
+constexpr char const *kBlobClassName = "RBlob";
 /// The class name of the RNTuple anchor
-static constexpr char const *kNTupleClassName = "ROOT::Experimental::RNTuple";
+constexpr char const *kNTupleClassName = "ROOT::Experimental::RNTuple";
 
 /// The RKeyBlob writes an invisible key into a TFile.  That is, a key that is not indexed in the list of keys,
 /// like a TBasket.
@@ -980,7 +1021,27 @@ ROOT::Experimental::Internal::RMiniFileReader::RMiniFileReader(ROOT::Internal::R
 {
 }
 
-ROOT::Experimental::RResult<ROOT::Experimental::Internal::RFileNTupleAnchor>
+ROOT::Experimental::RNTuple ROOT::Experimental::Internal::RMiniFileReader::CreateAnchor(
+   std::uint16_t versionEpoch, std::uint16_t versionMajor, std::uint16_t versionMinor, std::uint16_t versionPatch,
+   std::uint64_t seekHeader, std::uint64_t nbytesHeader, std::uint64_t lenHeader, std::uint64_t seekFooter,
+   std::uint64_t nbytesFooter, std::uint64_t lenFooter, std::uint64_t checksum)
+{
+   RNTuple ntuple;
+   ntuple.fVersionEpoch = versionEpoch;
+   ntuple.fVersionMajor = versionMajor;
+   ntuple.fVersionMinor = versionMinor;
+   ntuple.fVersionPatch = versionPatch;
+   ntuple.fSeekHeader = seekHeader;
+   ntuple.fNBytesHeader = nbytesHeader;
+   ntuple.fLenHeader = lenHeader;
+   ntuple.fSeekFooter = seekFooter;
+   ntuple.fNBytesFooter = nbytesFooter;
+   ntuple.fLenFooter = lenFooter;
+   ntuple.fChecksum = checksum;
+   return ntuple;
+}
+
+ROOT::Experimental::RResult<ROOT::Experimental::RNTuple>
 ROOT::Experimental::Internal::RMiniFileReader::GetNTuple(std::string_view ntupleName)
 {
    char ident[4];
@@ -991,7 +1052,7 @@ ROOT::Experimental::Internal::RMiniFileReader::GetNTuple(std::string_view ntuple
    return GetNTupleBare(ntupleName);
 }
 
-ROOT::Experimental::RResult<ROOT::Experimental::Internal::RFileNTupleAnchor>
+ROOT::Experimental::RResult<ROOT::Experimental::RNTuple>
 ROOT::Experimental::Internal::RMiniFileReader::GetNTupleProper(std::string_view ntupleName)
 {
    RTFHeader fileHeader;
@@ -1000,7 +1061,9 @@ ROOT::Experimental::Internal::RMiniFileReader::GetNTupleProper(std::string_view
    RTFKey key;
    RTFString name;
    ReadBuffer(&key, sizeof(key), fileHeader.fBEGIN);
+   // Skip over the entire key length, including the class name, object name, and title stored in it.
    std::uint64_t offset = fileHeader.fBEGIN + key.fKeyLen;
+   // Skip over the name and title of the TNamed preceding the TFile entry.
    ReadBuffer(&name, 1, offset);
    offset += name.GetSize();
    ReadBuffer(&name, 1, offset);
@@ -1042,19 +1105,40 @@ ROOT::Experimental::Internal::RMiniFileReader::GetNTupleProper(std::string_view
 
    ReadBuffer(&key, sizeof(key), key.GetSeekKey());
    offset = key.GetSeekKey() + key.fKeyLen;
-   RTFNTuple ntuple;
-   if (key.fObjLen > sizeof(ntuple)) {
-      return R__FAIL("invalid anchor size: " + std::to_string(key.fObjLen) + " > " + std::to_string(sizeof(ntuple)));
+
+   if (key.fObjLen < sizeof(RTFNTuple)) {
+      return R__FAIL("invalid anchor size: " + std::to_string(key.fObjLen) + " < " + std::to_string(sizeof(RTFNTuple)));
    }
-   ReadBuffer(&ntuple, key.fObjLen, offset);
-   if (sizeof(ntuple) != key.fObjLen) {
-      ROOT::Experimental::Detail::RNTupleDecompressor decompressor;
-      decompressor.Unzip(&ntuple, sizeof(ntuple), key.fObjLen);
+   // The object length can be larger than the size of RTFNTuple if it comes from a future RNTuple class version.
+   auto bufAnchor = std::make_unique<unsigned char[]>(key.fObjLen);
+   RTFNTuple *ntuple = new (bufAnchor.get()) RTFNTuple;
+
+   auto objNbytes = key.GetSize() - key.fKeyLen;
+   ReadBuffer(ntuple, objNbytes, offset);
+   if (objNbytes != key.fObjLen) {
+      RNTupleDecompressor decompressor;
+      decompressor.Unzip(bufAnchor.get(), objNbytes, key.fObjLen);
    }
-   return ntuple.ToAnchor();
+
+   if (ntuple->fVersionClass < 4) {
+      return R__FAIL("invalid anchor, unsupported pre-release of RNTuple");
+   }
+
+   // We require that future class versions only append members and store the checksum in the last 8 bytes
+   // Checksum calculation: strip byte count, class version, fChecksum member
+   RUInt64BE *ckOnDisk = reinterpret_cast<RUInt64BE *>(bufAnchor.get() + key.fObjLen - sizeof(RUInt64BE));
+   auto lenCkData = ntuple->GetSizeCkData() + key.fObjLen - sizeof(RTFNTuple);
+   auto ckCalc = XXH3_64bits(ntuple->GetPtrCkData(), lenCkData);
+   if (ckCalc != (uint64_t)(*ckOnDisk)) {
+      return R__FAIL("RNTuple anchor checksum mismatch");
+   }
+
+   return CreateAnchor(ntuple->fVersionEpoch, ntuple->fVersionMajor, ntuple->fVersionMinor, ntuple->fVersionPatch,
+                       ntuple->fSeekHeader, ntuple->fNBytesHeader, ntuple->fLenHeader, ntuple->fSeekFooter,
+                       ntuple->fNBytesFooter, ntuple->fLenFooter, ntuple->fChecksum);
 }
 
-ROOT::Experimental::RResult<ROOT::Experimental::Internal::RFileNTupleAnchor>
+ROOT::Experimental::RResult<ROOT::Experimental::RNTuple>
 ROOT::Experimental::Internal::RMiniFileReader::GetNTupleBare(std::string_view ntupleName)
 {
    RBareFileHeader fileHeader;
@@ -1073,7 +1157,12 @@ ROOT::Experimental::Internal::RMiniFileReader::GetNTupleBare(std::string_view nt
 
    RTFNTuple ntuple;
    ReadBuffer(&ntuple, sizeof(ntuple), offset);
-   return ntuple.ToAnchor();
+   auto checksum = XXH3_64bits(ntuple.GetPtrCkData(), ntuple.GetSizeCkData());
+   if (checksum != static_cast<uint64_t>(ntuple.fChecksum))
+      return R__FAIL("RNTuple bare file: anchor checksum mismatch");
+   return CreateAnchor(ntuple.fVersionEpoch, ntuple.fVersionMajor, ntuple.fVersionMinor, ntuple.fVersionPatch,
+                       ntuple.fSeekHeader, ntuple.fNBytesHeader, ntuple.fLenHeader, ntuple.fSeekFooter,
+                       ntuple.fNBytesFooter, ntuple.fLenFooter, ntuple.fChecksum);
 }
 
 
@@ -1105,11 +1194,13 @@ void ROOT::Experimental::Internal::RNTupleFileWriter::RFileSimple::Write(
 #else
       retval = fseek(fFile, offset, SEEK_SET);
 #endif
-      R__ASSERT(retval == 0);
+      if (retval)
+         throw RException(R__FAIL(std::string("Seek failed: ") + strerror(errno)));
       fFilePos = offset;
    }
    retval = fwrite(buffer, 1, nbytes, fFile);
-   R__ASSERT(retval == nbytes);
+   if (retval != nbytes)
+      throw RException(R__FAIL(std::string("write failed: ") + strerror(errno)));
    fFilePos += nbytes;
 }
 
@@ -1121,18 +1212,20 @@ std::uint64_t ROOT::Experimental::Internal::RNTupleFileWriter::RFileSimple::Writ
    const std::string &objectName,
    const std::string &title)
 {
-   if (offset < 0)
-      offset = fFilePos;
+   if (offset > 0)
+      fKeyOffset = offset;
    RTFString strClass{className};
    RTFString strObject{objectName};
    RTFString strTitle{title};
 
-   RTFKey key(offset, directoryOffset, strClass, strObject, strTitle, len, nbytes);
-   Write(&key, key.fKeyHeaderSize, offset);
+   RTFKey key(fKeyOffset, directoryOffset, strClass, strObject, strTitle, len, nbytes);
+   Write(&key, key.fKeyHeaderSize, fKeyOffset);
    Write(&strClass, strClass.GetSize());
    Write(&strObject, strObject.GetSize());
    Write(&strTitle, strTitle.GetSize());
    auto offsetData = fFilePos;
+   // The next key starts after the data.
+   fKeyOffset = offsetData + nbytes;
    if (buffer)
       Write(buffer, nbytes);
 
@@ -1149,7 +1242,8 @@ void ROOT::Experimental::Internal::RNTupleFileWriter::RFileProper::Write(
    R__ASSERT(fFile);
    fFile->Seek(offset);
    bool rv = fFile->WriteBuffer((char *)(buffer), nbytes);
-   R__ASSERT(!rv);
+   if (rv)
+      throw RException(R__FAIL("WriteBuffer failed."));
 }
 
 
@@ -1179,7 +1273,8 @@ std::uint64_t ROOT::Experimental::Internal::RNTupleFileWriter::RFileProper::Writ
    Write(&strTitle, strTitle.GetSize(), offset);
    offset += strTitle.GetSize();
    auto offsetData = offset;
-   Write(buffer, nbytes, offset);
+   if (buffer)
+      Write(buffer, nbytes, offset);
 
    return offsetData;
 }
@@ -1199,9 +1294,9 @@ ROOT::Experimental::Internal::RNTupleFileWriter::~RNTupleFileWriter()
 {
 }
 
-
-ROOT::Experimental::Internal::RNTupleFileWriter *ROOT::Experimental::Internal::RNTupleFileWriter::Recreate(
-   std::string_view ntupleName, std::string_view path, int defaultCompression, ENTupleContainerFormat containerFormat)
+ROOT::Experimental::Internal::RNTupleFileWriter *
+ROOT::Experimental::Internal::RNTupleFileWriter::Recreate(std::string_view ntupleName, std::string_view path,
+                                                          int defaultCompression, EContainerFormat containerFormat)
 {
    std::string fileName(path);
    size_t idxDirSep = fileName.find_last_of("\\/");
@@ -1220,10 +1315,8 @@ ROOT::Experimental::Internal::RNTupleFileWriter *ROOT::Experimental::Internal::R
    writer->fFileName = fileName;
 
    switch (containerFormat) {
-   case ENTupleContainerFormat::kTFile:
-      writer->WriteTFileSkeleton(defaultCompression);
-      break;
-   case ENTupleContainerFormat::kBare:
+   case EContainerFormat::kTFile: writer->WriteTFileSkeleton(defaultCompression); break;
+   case EContainerFormat::kBare:
       writer->fIsBare = true;
       writer->WriteBareFileSkeleton(defaultCompression);
       break;
@@ -1235,18 +1328,6 @@ ROOT::Experimental::Internal::RNTupleFileWriter *ROOT::Experimental::Internal::R
 }
 
 
-ROOT::Experimental::Internal::RNTupleFileWriter *ROOT::Experimental::Internal::RNTupleFileWriter::Recreate(
-   std::string_view ntupleName, std::string_view path, std::unique_ptr<TFile> &file)
-{
-   file = std::unique_ptr<TFile>(TFile::Open(std::string(path.data(), path.size()).c_str(), "RECREATE"));
-   R__ASSERT(file && !file->IsZombie());
-
-   auto writer = new RNTupleFileWriter(ntupleName);
-   writer->fFileProper.fFile = file.get();
-   return writer;
-}
-
-
 ROOT::Experimental::Internal::RNTupleFileWriter *ROOT::Experimental::Internal::RNTupleFileWriter::Append(
    std::string_view ntupleName, TFile &file)
 {
@@ -1260,8 +1341,7 @@ void ROOT::Experimental::Internal::RNTupleFileWriter::Commit()
 {
    if (fFileProper) {
       // Easy case, the ROOT file header and the RNTuple streaming is taken care of by TFile
-      ROOT::Experimental::RNTuple ntuple(fNTupleAnchor);
-      fFileProper.fFile->WriteObject(&ntuple, fNTupleName.c_str());
+      fFileProper.fFile->WriteObject(&fNTupleAnchor, fNTupleName.c_str());
       fFileProper.fFile->Write();
       return;
    }
@@ -1294,8 +1374,9 @@ std::uint64_t ROOT::Experimental::Internal::RNTupleFileWriter::WriteBlob(const v
    std::uint64_t offset;
    if (fFileSimple) {
       if (fIsBare) {
-         offset = fFileSimple.fFilePos;
+         offset = fFileSimple.fKeyOffset;
          fFileSimple.Write(data, nbytes);
+         fFileSimple.fKeyOffset += nbytes;
       } else {
          offset = fFileSimple.WriteKey(data, nbytes, len, -1, 100, kBlobClassName);
       }
@@ -1305,6 +1386,31 @@ std::uint64_t ROOT::Experimental::Internal::RNTupleFileWriter::WriteBlob(const v
    return offset;
 }
 
+std::uint64_t ROOT::Experimental::Internal::RNTupleFileWriter::ReserveBlob(size_t nbytes, size_t len)
+{
+   std::uint64_t offset;
+   if (fFileSimple) {
+      if (fIsBare) {
+         offset = fFileSimple.fKeyOffset;
+         fFileSimple.fKeyOffset += nbytes;
+      } else {
+         offset = fFileSimple.WriteKey(/*buffer=*/nullptr, nbytes, len, -1, 100, kBlobClassName);
+      }
+   } else {
+      offset = fFileProper.WriteKey(/*buffer=*/nullptr, nbytes, len);
+   }
+   return offset;
+}
+
+void ROOT::Experimental::Internal::RNTupleFileWriter::WriteIntoReservedBlob(const void *buffer, size_t nbytes,
+                                                                            std::int64_t offset)
+{
+   if (fFileSimple) {
+      fFileSimple.Write(buffer, nbytes, offset);
+   } else {
+      fFileProper.Write(buffer, nbytes, offset);
+   }
+}
 
 std::uint64_t ROOT::Experimental::Internal::RNTupleFileWriter::WriteNTupleHeader(
    const void *data, size_t nbytes, size_t lenHeader)
@@ -1340,6 +1446,7 @@ void ROOT::Experimental::Internal::RNTupleFileWriter::WriteBareFileSkeleton(int
    RTFNTuple ntupleOnDisk;
    fFileSimple.fControlBlock->fSeekNTuple = fFileSimple.fFilePos;
    fFileSimple.Write(&ntupleOnDisk, ntupleOnDisk.GetSize());
+   fFileSimple.fKeyOffset = fFileSimple.fFilePos;
 }
 
 void ROOT::Experimental::Internal::RNTupleFileWriter::WriteTFileStreamerInfo()
@@ -1348,23 +1455,24 @@ void ROOT::Experimental::Internal::RNTupleFileWriter::WriteTFileStreamerInfo()
    RTFString strStreamerInfo{"StreamerInfo"};
    RTFString strStreamerTitle{"Doubly linked list"};
 
-   fFileSimple.fControlBlock->fHeader.SetSeekInfo(fFileSimple.fFilePos);
+   fFileSimple.fControlBlock->fHeader.SetSeekInfo(fFileSimple.fKeyOffset);
    RTFKey keyStreamerInfo(
       fFileSimple.fControlBlock->fHeader.GetSeekInfo(), 100, strTList, strStreamerInfo, strStreamerTitle, 0);
    RTFStreamerInfoList streamerInfo;
    auto classTagOffset = keyStreamerInfo.fKeyLen + offsetof(struct RTFStreamerInfoList, fStreamerInfo) +
                          offsetof(struct RTFStreamerInfoObject, fStreamers) +
-                         offsetof(struct RTFStreamerChecksum, fNewClassTag) + 2;
-   streamerInfo.fStreamerInfo.fStreamers.fStreamerVersion.fClassTag = 0x80000000 | classTagOffset;
-   streamerInfo.fStreamerInfo.fStreamers.fStreamerSize.fClassTag = 0x80000000 | classTagOffset;
+                         offsetof(struct RTFStreamerVersionEpoch, fNewClassTag) + 2;
+   streamerInfo.fStreamerInfo.fStreamers.fStreamerVersionMajor.fClassTag = 0x80000000 | classTagOffset;
+   streamerInfo.fStreamerInfo.fStreamers.fStreamerVersionMinor.fClassTag = 0x80000000 | classTagOffset;
+   streamerInfo.fStreamerInfo.fStreamers.fStreamerVersionPatch.fClassTag = 0x80000000 | classTagOffset;
    streamerInfo.fStreamerInfo.fStreamers.fStreamerSeekHeader.fClassTag = 0x80000000 | classTagOffset;
    streamerInfo.fStreamerInfo.fStreamers.fStreamerNBytesHeader.fClassTag = 0x80000000 | classTagOffset;
    streamerInfo.fStreamerInfo.fStreamers.fStreamerLenHeader.fClassTag = 0x80000000 | classTagOffset;
    streamerInfo.fStreamerInfo.fStreamers.fStreamerSeekFooter.fClassTag = 0x80000000 | classTagOffset;
    streamerInfo.fStreamerInfo.fStreamers.fStreamerNBytesFooter.fClassTag = 0x80000000 | classTagOffset;
    streamerInfo.fStreamerInfo.fStreamers.fStreamerLenFooter.fClassTag = 0x80000000 | classTagOffset;
-   streamerInfo.fStreamerInfo.fStreamers.fStreamerReserved.fClassTag = 0x80000000 | classTagOffset;
-   Detail::RNTupleCompressor compressor;
+   streamerInfo.fStreamerInfo.fStreamers.fStreamerChecksum.fClassTag = 0x80000000 | classTagOffset;
+   RNTupleCompressor compressor;
    auto szStreamerInfo = compressor.Zip(&streamerInfo, streamerInfo.GetSize(), 1);
    fFileSimple.WriteKey(compressor.GetZipBuffer(), szStreamerInfo, streamerInfo.GetSize(),
                         fFileSimple.fControlBlock->fHeader.GetSeekInfo(), 100,
@@ -1383,7 +1491,7 @@ void ROOT::Experimental::Internal::RNTupleFileWriter::WriteTFileKeysList()
    RTFKey keyRNTuple(fFileSimple.fControlBlock->fSeekNTuple, 100, strRNTupleClass, strRNTupleName, strEmpty,
                      RTFNTuple().GetSize());
 
-   fFileSimple.fControlBlock->fFileRecord.SetSeekKeys(fFileSimple.fFilePos);
+   fFileSimple.fControlBlock->fFileRecord.SetSeekKeys(fFileSimple.fKeyOffset);
    RTFKeyList keyList{1};
    RTFKey keyKeyList(fFileSimple.fControlBlock->fFileRecord.GetSeekKeys(), 100, strEmpty, strFileName, strEmpty,
                      keyList.GetSize() + keyRNTuple.fKeyLen);
@@ -1393,16 +1501,18 @@ void ROOT::Experimental::Internal::RNTupleFileWriter::WriteTFileKeysList()
    fFileSimple.Write(&strEmpty, strEmpty.GetSize());
    fFileSimple.Write(&keyList, keyList.GetSize());
    fFileSimple.Write(&keyRNTuple, keyRNTuple.fKeyHeaderSize);
+   // Write class name, object name, and title for this key.
    fFileSimple.Write(&strRNTupleClass, strRNTupleClass.GetSize());
    fFileSimple.Write(&strRNTupleName, strRNTupleName.GetSize());
    fFileSimple.Write(&strEmpty, strEmpty.GetSize());
    fFileSimple.fControlBlock->fFileRecord.fNBytesKeys =
       fFileSimple.fFilePos - fFileSimple.fControlBlock->fFileRecord.GetSeekKeys();
+   fFileSimple.fKeyOffset = fFileSimple.fFilePos;
 }
 
 void ROOT::Experimental::Internal::RNTupleFileWriter::WriteTFileFreeList()
 {
-   fFileSimple.fControlBlock->fHeader.SetSeekFree(fFileSimple.fFilePos);
+   fFileSimple.fControlBlock->fHeader.SetSeekFree(fFileSimple.fKeyOffset);
    RTFString strEmpty;
    RTFString strFileName{fFileName};
    RTFFreeEntry freeEntry;
@@ -1424,7 +1534,7 @@ void ROOT::Experimental::Internal::RNTupleFileWriter::WriteTFileNTupleKey()
    RTFString strEmpty;
 
    RTFNTuple ntupleOnDisk(fNTupleAnchor);
-   fFileSimple.fControlBlock->fSeekNTuple = fFileSimple.fFilePos;
+   fFileSimple.fControlBlock->fSeekNTuple = fFileSimple.fKeyOffset;
    fFileSimple.WriteKey(&ntupleOnDisk, ntupleOnDisk.GetSize(), ntupleOnDisk.GetSize(),
                         fFileSimple.fControlBlock->fSeekNTuple, 100, "ROOT::Experimental::RNTuple", fNTupleName, "");
 }
@@ -1447,9 +1557,11 @@ void ROOT::Experimental::Internal::RNTupleFileWriter::WriteTFileSkeleton(int def
    fFileSimple.fControlBlock->fHeader.SetNbytesName(nbytesName);
 
    fFileSimple.Write(&keyRoot, keyRoot.fKeyHeaderSize, 100);
+   // Write class name, object name, and title for the TFile key.
    fFileSimple.Write(&strTFile, strTFile.GetSize());
    fFileSimple.Write(&strFileName, strFileName.GetSize());
    fFileSimple.Write(&strEmpty, strEmpty.GetSize());
+   // Write the name and title of the TNamed preceding the TFile entry.
    fFileSimple.Write(&strFileName, strFileName.GetSize());
    fFileSimple.Write(&strEmpty, strEmpty.GetSize());
    // Will be overwritten on commit
@@ -1461,4 +1573,5 @@ void ROOT::Experimental::Internal::RNTupleFileWriter::WriteTFileSkeleton(int def
    RUInt32BE padding{0};
    for (int i = 0; i < 3; ++i)
       fFileSimple.Write(&padding, sizeof(padding));
+   fFileSimple.fKeyOffset = fFileSimple.fFilePos;
 }
