@@ -102,59 +102,65 @@ TEST(RNTuple, SerializeFieldStructure)
 TEST(RNTuple, SerializeEnvelope)
 {
    try {
-      RNTupleSerializer::DeserializeEnvelope(nullptr, 0).Unwrap();
+      RNTupleSerializer::DeserializeEnvelope(nullptr, 0, 137).Unwrap();
       FAIL() << "too small envelope should throw";
    } catch (const RException& err) {
       EXPECT_THAT(err.what(), testing::HasSubstr("too short"));
    }
 
    struct {
-      std::uint16_t writerVersion = static_cast<std::uint16_t>(1);
-      std::uint16_t minVersion = static_cast<std::uint16_t>(1);
-      std::uint32_t crc32 = 0;
+      std::uint64_t typeAndSize = 137;
+      std::uint64_t payload = 0;
+      std::uint64_t xxhash3 = 0;
    } testEnvelope;
 
+   EXPECT_EQ(8u, RNTupleSerializer::SerializeEnvelopePostscript(reinterpret_cast<unsigned char *>(&testEnvelope), 16));
+   testEnvelope.xxhash3 = 0;
    try {
-      RNTupleSerializer::DeserializeEnvelope(&testEnvelope, sizeof(testEnvelope)).Unwrap();
-      FAIL() << "CRC32 mismatch should throw";
+      RNTupleSerializer::DeserializeEnvelope(&testEnvelope, sizeof(testEnvelope), 137).Unwrap();
+      FAIL() << "XxHash-3 mismatch should throw";
    } catch (const RException& err) {
-      EXPECT_THAT(err.what(), testing::HasSubstr("CRC32"));
+      EXPECT_THAT(err.what(), testing::HasSubstr("XxHash-3"));
+   }
+   testEnvelope.typeAndSize = 137;
+
+   EXPECT_EQ(8u, RNTupleSerializer::SerializeEnvelopePostscript(reinterpret_cast<unsigned char *>(&testEnvelope), 16));
+   try {
+      RNTupleSerializer::DeserializeEnvelope(&testEnvelope, sizeof(testEnvelope), 138).Unwrap();
+      FAIL() << "unsupported envelope type should throw";
+   } catch (const RException &err) {
+      EXPECT_THAT(err.what(), testing::HasSubstr("envelope type mismatch"));
+   }
+   testEnvelope.typeAndSize = 137;
+
+   EXPECT_EQ(8u, RNTupleSerializer::SerializeEnvelopePostscript(reinterpret_cast<unsigned char *>(&testEnvelope), 16));
+   try {
+      RNTupleSerializer::DeserializeEnvelope(&testEnvelope, sizeof(testEnvelope) - 1, 137).Unwrap();
+      FAIL() << "too small envelope buffer should throw";
+   } catch (const RException& err) {
+      EXPECT_THAT(err.what(), testing::HasSubstr("envelope buffer size too small"));
    }
 
-   testEnvelope.writerVersion = testEnvelope.minVersion = 0;
-
-   EXPECT_EQ(4u, RNTupleSerializer::SerializeEnvelopePostscript(
-      reinterpret_cast<const unsigned char *>(&testEnvelope), 4, &testEnvelope.crc32));
+   testEnvelope.typeAndSize -= 9 << 16;
    try {
-      RNTupleSerializer::DeserializeEnvelope(&testEnvelope, sizeof(testEnvelope)).Unwrap();
-      FAIL() << "unsupported version should throw";
+      RNTupleSerializer::DeserializeEnvelope(&testEnvelope, sizeof(testEnvelope), 137).Unwrap();
+      FAIL() << "too small envelope buffer should throw";
    } catch (const RException& err) {
-      EXPECT_THAT(err.what(), testing::HasSubstr("too old"));
+      EXPECT_THAT(err.what(), testing::HasSubstr("invalid envelope, too short"));
    }
 
-   testEnvelope.writerVersion = RNTupleSerializer::kEnvelopeCurrentVersion;
-   std::uint32_t crc32_write;
-   RNTupleSerializer::SerializeEnvelopePostscript(
-      reinterpret_cast<const unsigned char *>(&testEnvelope), 4, crc32_write, &testEnvelope.crc32);
-   std::uint32_t crc32_read;
-   EXPECT_EQ(4u, RNTupleSerializer::DeserializeEnvelope(&testEnvelope, sizeof(testEnvelope), crc32_read).Unwrap());
-   EXPECT_EQ(crc32_write, crc32_read);
-
-   testEnvelope.writerVersion = RNTupleSerializer::kEnvelopeCurrentVersion + 1;
-   testEnvelope.minVersion = RNTupleSerializer::kEnvelopeCurrentVersion + 1;
-   EXPECT_EQ(4u, RNTupleSerializer::SerializeEnvelopePostscript(
-      reinterpret_cast<const unsigned char *>(&testEnvelope), 4, &testEnvelope.crc32));
-   try {
-      RNTupleSerializer::DeserializeEnvelope(&testEnvelope, sizeof(testEnvelope)).Unwrap();
-      FAIL() << "unsupported version should throw";
-   } catch (const RException& err) {
-      EXPECT_THAT(err.what(), testing::HasSubstr("too new"));
-   }
+   testEnvelope.typeAndSize = 137;
+   std::uint64_t xxhash3_write;
+   RNTupleSerializer::SerializeEnvelopePostscript(reinterpret_cast<unsigned char *>(&testEnvelope), 16, xxhash3_write);
+   std::uint64_t xxhash3_read;
+   EXPECT_EQ(8u,
+             RNTupleSerializer::DeserializeEnvelope(&testEnvelope, sizeof(testEnvelope), 137, xxhash3_read).Unwrap());
+   EXPECT_EQ(xxhash3_write, xxhash3_read);
 }
 
 TEST(RNTuple, SerializeFrame)
 {
-   std::uint32_t frameSize;
+   std::uint64_t frameSize;
    std::uint32_t nitems;
 
    try {
@@ -164,63 +170,74 @@ TEST(RNTuple, SerializeFrame)
       EXPECT_THAT(err.what(), testing::HasSubstr("too short"));
    }
 
-   unsigned char buffer[12];
-   EXPECT_EQ(4u, RNTupleSerializer::SerializeRecordFramePreamble(buffer));
-   try {
-      RNTupleSerializer::SerializeFramePostscript(buffer, 3);
-      FAIL() << "too small frame should throw";
-   } catch (const RException& err) {
-      EXPECT_THAT(err.what(), testing::HasSubstr("too short"));
-   }
-   EXPECT_EQ(0u, RNTupleSerializer::SerializeFramePostscript(buffer, 6));
-
-   try {
-      RNTupleSerializer::DeserializeFrameHeader(buffer, 4, frameSize).Unwrap();
-      FAIL() << "too small frame should throw";
-   } catch (const RException& err) {
-      EXPECT_THAT(err.what(), testing::HasSubstr("too short"));
-   }
-   EXPECT_EQ(4u, RNTupleSerializer::DeserializeFrameHeader(buffer, 6, frameSize, nitems).Unwrap());
-   EXPECT_EQ(6u, frameSize);
-   EXPECT_EQ(1u, nitems);
-
-   EXPECT_EQ(4u, RNTupleSerializer::SerializeRecordFramePreamble(buffer));
-   try {
-      RNTupleSerializer::SerializeFramePostscript(buffer, -2);
-      FAIL() << "too big frame should throw";
-   } catch (const RException& err) {
-      EXPECT_THAT(err.what(), testing::HasSubstr("too large"));
-   }
-
-   try {
-      RNTupleSerializer::SerializeListFramePreamble(1 << 29, buffer);
-      FAIL() << "too big list frame should throw";
-   } catch (const RException& err) {
-      EXPECT_THAT(err.what(), testing::HasSubstr("too large"));
-   }
-   EXPECT_EQ(8u, RNTupleSerializer::SerializeListFramePreamble(2, buffer));
+   unsigned char buffer[14];
+   EXPECT_EQ(8u, RNTupleSerializer::SerializeRecordFramePreamble(buffer));
    try {
       RNTupleSerializer::SerializeFramePostscript(buffer, 7);
       FAIL() << "too small frame should throw";
    } catch (const RException& err) {
       EXPECT_THAT(err.what(), testing::HasSubstr("too short"));
    }
-   EXPECT_EQ(0u, RNTupleSerializer::SerializeFramePostscript(buffer, 12));
+   EXPECT_EQ(0u, RNTupleSerializer::SerializeFramePostscript(buffer, 10));
 
    try {
-      RNTupleSerializer::DeserializeFrameHeader(buffer, 6, frameSize, nitems).Unwrap();
+      RNTupleSerializer::DeserializeFrameHeader(buffer, 8, frameSize).Unwrap();
+      FAIL() << "too small frame should throw";
+   } catch (const RException& err) {
+      EXPECT_THAT(err.what(), testing::HasSubstr("too short"));
+   }
+   EXPECT_EQ(8u, RNTupleSerializer::DeserializeFrameHeader(buffer, 10, frameSize, nitems).Unwrap());
+   EXPECT_EQ(10u, frameSize);
+   EXPECT_EQ(1u, nitems);
+
+   EXPECT_EQ(12u, RNTupleSerializer::SerializeListFramePreamble(2, buffer));
+   try {
+      RNTupleSerializer::SerializeFramePostscript(buffer, 11);
+      FAIL() << "too small frame should throw";
+   } catch (const RException& err) {
+      EXPECT_THAT(err.what(), testing::HasSubstr("too short"));
+   }
+   EXPECT_EQ(0u, RNTupleSerializer::SerializeFramePostscript(buffer, 14));
+
+   try {
+      RNTupleSerializer::DeserializeFrameHeader(buffer, 11, frameSize, nitems).Unwrap();
       FAIL() << "too short list frame should throw";
    } catch (const RException& err) {
       EXPECT_THAT(err.what(), testing::HasSubstr("too short"));
    }
-   EXPECT_EQ(8u, RNTupleSerializer::DeserializeFrameHeader(buffer, 12, frameSize, nitems).Unwrap());
-   EXPECT_EQ(12u, frameSize);
+   EXPECT_EQ(12u, RNTupleSerializer::DeserializeFrameHeader(buffer, 14, frameSize, nitems).Unwrap());
+   EXPECT_EQ(14u, frameSize);
    EXPECT_EQ(2u, nitems);
 }
 
+TEST(RNTuple, SerializeLongFrame)
+{
+   std::uint64_t frameSize;
+   std::uint32_t nitems;
+
+   const std::size_t payloadSize = 1024 * 1024 * 1024;
+   const std::size_t bufSize = payloadSize + 12;
+
+   auto buffer = std::make_unique<char[]>(bufSize);
+   buffer[12] = 'x';
+   EXPECT_EQ(12u, RNTupleSerializer::SerializeListFramePreamble(2, buffer.get()));
+   EXPECT_EQ(0u, RNTupleSerializer::SerializeFramePostscript(buffer.get(), payloadSize + 12));
+
+   try {
+      RNTupleSerializer::DeserializeFrameHeader(buffer.get(), payloadSize, frameSize).Unwrap();
+      FAIL() << "too small frame should throw";
+   } catch (const RException &err) {
+      EXPECT_THAT(err.what(), testing::HasSubstr("too short"));
+   }
+   EXPECT_EQ(12u, RNTupleSerializer::DeserializeFrameHeader(buffer.get(), bufSize, frameSize, nitems).Unwrap());
+   EXPECT_EQ(bufSize, frameSize);
+   EXPECT_EQ(2u, nitems);
+   EXPECT_EQ('x', buffer[12]);
+}
+
 TEST(RNTuple, SerializeFeatureFlags)
 {
-   std::vector<std::int64_t> flags;
+   std::vector<std::uint64_t> flags;
    unsigned char buffer[16];
 
    EXPECT_EQ(8u, RNTupleSerializer::SerializeFeatureFlags(flags, nullptr));
@@ -235,7 +252,7 @@ TEST(RNTuple, SerializeFeatureFlags)
    ASSERT_EQ(1u, flags.size());
    EXPECT_EQ(1, flags[0]);
 
-   flags.push_back(-2);
+   flags.push_back(std::uint64_t(-2));
    try {
       RNTupleSerializer::SerializeFeatureFlags(flags, buffer);
       FAIL() << "negative feature flag should throw";
@@ -347,13 +364,13 @@ TEST(RNTuple, SerializeLocator)
 TEST(RNTuple, SerializeEnvelopeLink)
 {
    RNTupleSerializer::REnvelopeLink link;
-   link.fUnzippedSize = 42;
+   link.fLength = 42;
    link.fLocator.fPosition = 137U;
    link.fLocator.fBytesOnStorage = 7;
 
-   unsigned char buffer[16];
-   EXPECT_EQ(16u, RNTupleSerializer::SerializeEnvelopeLink(link, nullptr));
-   EXPECT_EQ(16u, RNTupleSerializer::SerializeEnvelopeLink(link, buffer));
+   unsigned char buffer[20];
+   EXPECT_EQ(20u, RNTupleSerializer::SerializeEnvelopeLink(link, nullptr));
+   EXPECT_EQ(20u, RNTupleSerializer::SerializeEnvelopeLink(link, buffer));
    try {
       RNTupleSerializer::DeserializeEnvelopeLink(buffer, 3, link).Unwrap();
       FAIL() << "too short envelope link buffer should throw";
@@ -362,8 +379,8 @@ TEST(RNTuple, SerializeEnvelopeLink)
    }
 
    RNTupleSerializer::REnvelopeLink reconstructedLink;
-   EXPECT_EQ(16u, RNTupleSerializer::DeserializeEnvelopeLink(buffer, 16, reconstructedLink).Unwrap());
-   EXPECT_EQ(link.fUnzippedSize, reconstructedLink.fUnzippedSize);
+   EXPECT_EQ(20u, RNTupleSerializer::DeserializeEnvelopeLink(buffer, 20, reconstructedLink).Unwrap());
+   EXPECT_EQ(link.fLength, reconstructedLink.fLength);
    EXPECT_EQ(link.fLocator, reconstructedLink.fLocator);
 }
 
@@ -373,24 +390,10 @@ TEST(RNTuple, SerializeClusterSummary)
    summary.fFirstEntry = 42;
    summary.fNEntries = 137;
 
-   unsigned char buffer[24];
-   ASSERT_EQ(20u, RNTupleSerializer::SerializeClusterSummary(summary, nullptr));
-   EXPECT_EQ(20u, RNTupleSerializer::SerializeClusterSummary(summary, buffer));
-   RNTupleSerializer::RClusterSummary reco;
-   try {
-      RNTupleSerializer::DeserializeClusterSummary(buffer, 19, reco).Unwrap();
-      FAIL() << "too short cluster summary should fail";
-   } catch (const RException& err) {
-      EXPECT_THAT(err.what(), testing::HasSubstr("too short"));
-   }
-   EXPECT_EQ(20u, RNTupleSerializer::DeserializeClusterSummary(buffer, 20, reco).Unwrap());
-   EXPECT_EQ(summary.fFirstEntry, reco.fFirstEntry);
-   EXPECT_EQ(summary.fNEntries, reco.fNEntries);
-   EXPECT_EQ(summary.fColumnGroupID, reco.fColumnGroupID);
-
-   summary.fColumnGroupID = 13;
+   unsigned char buffer[28];
    ASSERT_EQ(24u, RNTupleSerializer::SerializeClusterSummary(summary, nullptr));
    EXPECT_EQ(24u, RNTupleSerializer::SerializeClusterSummary(summary, buffer));
+   RNTupleSerializer::RClusterSummary reco;
    try {
       RNTupleSerializer::DeserializeClusterSummary(buffer, 23, reco).Unwrap();
       FAIL() << "too short cluster summary should fail";
@@ -401,29 +404,47 @@ TEST(RNTuple, SerializeClusterSummary)
    EXPECT_EQ(summary.fFirstEntry, reco.fFirstEntry);
    EXPECT_EQ(summary.fNEntries, reco.fNEntries);
    EXPECT_EQ(summary.fColumnGroupID, reco.fColumnGroupID);
+
+   summary.fColumnGroupID = 13;
+   ASSERT_EQ(28u, RNTupleSerializer::SerializeClusterSummary(summary, nullptr));
+   EXPECT_EQ(28u, RNTupleSerializer::SerializeClusterSummary(summary, buffer));
+   try {
+      RNTupleSerializer::DeserializeClusterSummary(buffer, 27, reco).Unwrap();
+      FAIL() << "too short cluster summary should fail";
+   } catch (const RException& err) {
+      EXPECT_THAT(err.what(), testing::HasSubstr("too short"));
+   }
+   EXPECT_EQ(28u, RNTupleSerializer::DeserializeClusterSummary(buffer, 28, reco).Unwrap());
+   EXPECT_EQ(summary.fFirstEntry, reco.fFirstEntry);
+   EXPECT_EQ(summary.fNEntries, reco.fNEntries);
+   EXPECT_EQ(summary.fColumnGroupID, reco.fColumnGroupID);
 }
 
 TEST(RNTuple, SerializeClusterGroup)
 {
    RNTupleSerializer::RClusterGroup group;
+   group.fMinEntry = 7;
+   group.fEntrySpan = 84600;
    group.fNClusters = 42;
-   group.fPageListEnvelopeLink.fUnzippedSize = 42;
+   group.fPageListEnvelopeLink.fLength = 42;
    group.fPageListEnvelopeLink.fLocator.fPosition = 137U;
    group.fPageListEnvelopeLink.fLocator.fBytesOnStorage = 7;
 
-   unsigned char buffer[28];
-   ASSERT_EQ(24u, RNTupleSerializer::SerializeClusterGroup(group, nullptr));
-   EXPECT_EQ(24u, RNTupleSerializer::SerializeClusterGroup(group, buffer));
+   unsigned char buffer[52];
+   ASSERT_EQ(48u, RNTupleSerializer::SerializeClusterGroup(group, nullptr));
+   EXPECT_EQ(48u, RNTupleSerializer::SerializeClusterGroup(group, buffer));
    RNTupleSerializer::RClusterGroup reco;
    try {
-      RNTupleSerializer::DeserializeClusterGroup(buffer, 23, reco).Unwrap();
+      RNTupleSerializer::DeserializeClusterGroup(buffer, 47, reco).Unwrap();
       FAIL() << "too short cluster group should fail";
    } catch (const RException& err) {
       EXPECT_THAT(err.what(), testing::HasSubstr("too short"));
    }
-   EXPECT_EQ(24u, RNTupleSerializer::DeserializeClusterGroup(buffer, 24, reco).Unwrap());
+   EXPECT_EQ(48u, RNTupleSerializer::DeserializeClusterGroup(buffer, 48, reco).Unwrap());
+   EXPECT_EQ(group.fMinEntry, reco.fMinEntry);
+   EXPECT_EQ(group.fEntrySpan, reco.fEntrySpan);
    EXPECT_EQ(group.fNClusters, reco.fNClusters);
-   EXPECT_EQ(group.fPageListEnvelopeLink.fUnzippedSize, reco.fPageListEnvelopeLink.fUnzippedSize);
+   EXPECT_EQ(group.fPageListEnvelopeLink.fLength, reco.fPageListEnvelopeLink.fLength);
    EXPECT_EQ(group.fPageListEnvelopeLink.fLocator.GetPosition<std::uint64_t>(),
              reco.fPageListEnvelopeLink.fLocator.GetPosition<std::uint64_t>());
    EXPECT_EQ(group.fPageListEnvelopeLink.fLocator.fBytesOnStorage, reco.fPageListEnvelopeLink.fLocator.fBytesOnStorage);
@@ -431,21 +452,25 @@ TEST(RNTuple, SerializeClusterGroup)
    // Test frame evolution
    auto pos = buffer;
    pos += RNTupleSerializer::SerializeRecordFramePreamble(pos);
+   pos += RNTupleSerializer::SerializeUInt64(group.fMinEntry, pos);
+   pos += RNTupleSerializer::SerializeUInt64(group.fEntrySpan, pos);
    pos += RNTupleSerializer::SerializeUInt32(group.fNClusters, pos);
    pos += RNTupleSerializer::SerializeEnvelopeLink(group.fPageListEnvelopeLink, pos);
    pos += RNTupleSerializer::SerializeUInt16(7, pos);
    pos += RNTupleSerializer::SerializeFramePostscript(buffer, pos - buffer);
    pos += RNTupleSerializer::SerializeUInt16(13, pos);
-   EXPECT_EQ(26u, RNTupleSerializer::DeserializeClusterGroup(buffer, 26, reco).Unwrap());
+   EXPECT_EQ(50u, RNTupleSerializer::DeserializeClusterGroup(buffer, 50, reco).Unwrap());
+   EXPECT_EQ(group.fMinEntry, reco.fMinEntry);
+   EXPECT_EQ(group.fEntrySpan, reco.fEntrySpan);
    EXPECT_EQ(group.fNClusters, reco.fNClusters);
-   EXPECT_EQ(group.fPageListEnvelopeLink.fUnzippedSize, reco.fPageListEnvelopeLink.fUnzippedSize);
+   EXPECT_EQ(group.fPageListEnvelopeLink.fLength, reco.fPageListEnvelopeLink.fLength);
    EXPECT_EQ(group.fPageListEnvelopeLink.fLocator.GetPosition<std::uint64_t>(),
              reco.fPageListEnvelopeLink.fLocator.GetPosition<std::uint64_t>());
    EXPECT_EQ(group.fPageListEnvelopeLink.fLocator.fBytesOnStorage, reco.fPageListEnvelopeLink.fLocator.fBytesOnStorage);
    std::uint16_t remainder;
-   RNTupleSerializer::DeserializeUInt16(buffer + 24, remainder);
+   RNTupleSerializer::DeserializeUInt16(buffer + 48, remainder);
    EXPECT_EQ(7u, remainder);
-   RNTupleSerializer::DeserializeUInt16(buffer + 26, remainder);
+   RNTupleSerializer::DeserializeUInt16(buffer + 50, remainder);
    EXPECT_EQ(13u, remainder);
 }
 
@@ -460,12 +485,12 @@ TEST(RNTuple, SerializeEmptyHeader)
       .MakeDescriptor()
       .Unwrap());
    auto desc = builder.MoveDescriptor();
-   auto context = RNTupleSerializer::SerializeHeaderV1(nullptr, desc);
+   auto context = RNTupleSerializer::SerializeHeader(nullptr, desc);
    EXPECT_GT(context.GetHeaderSize(), 0);
    auto buffer = std::make_unique<unsigned char []>(context.GetHeaderSize());
-   context = RNTupleSerializer::SerializeHeaderV1(buffer.get(), desc);
+   context = RNTupleSerializer::SerializeHeader(buffer.get(), desc);
 
-   RNTupleSerializer::DeserializeHeaderV1(buffer.get(), context.GetHeaderSize(), builder);
+   RNTupleSerializer::DeserializeHeader(buffer.get(), context.GetHeaderSize(), builder);
 }
 
 
@@ -513,12 +538,12 @@ TEST(RNTuple, SerializeHeader)
    builder.AddColumn(40, 40, 137, RColumnModel(EColumnType::kByte, true), 1);
 
    auto desc = builder.MoveDescriptor();
-   auto context = RNTupleSerializer::SerializeHeaderV1(nullptr, desc);
+   auto context = RNTupleSerializer::SerializeHeader(nullptr, desc);
    EXPECT_GT(context.GetHeaderSize(), 0);
    auto buffer = std::make_unique<unsigned char []>(context.GetHeaderSize());
-   context = RNTupleSerializer::SerializeHeaderV1(buffer.get(), desc);
+   context = RNTupleSerializer::SerializeHeader(buffer.get(), desc);
 
-   RNTupleSerializer::DeserializeHeaderV1(buffer.get(), context.GetHeaderSize(), builder);
+   RNTupleSerializer::DeserializeHeader(buffer.get(), context.GetHeaderSize(), builder);
 
    desc = builder.MoveDescriptor();
    auto ptAliasFieldId = desc.FindFieldId("ptAlias");
@@ -550,27 +575,29 @@ TEST(RNTuple, SerializeFooter)
 
    ROOT::Experimental::RClusterDescriptor::RColumnRange columnRange;
    ROOT::Experimental::RClusterDescriptor::RPageRange::RPageInfo pageInfo;
-   RClusterDescriptorBuilder clusterBuilder(84, 0, 100);
+   RClusterDescriptorBuilder clusterBuilder;
+   clusterBuilder.ClusterId(84).FirstEntryIndex(0).NEntries(100);
    ROOT::Experimental::RClusterDescriptor::RPageRange pageRange;
    pageRange.fPhysicalColumnId = 17;
    pageInfo.fNElements = 100;
    pageInfo.fLocator.fPosition = 7000U;
    pageRange.fPageInfos.emplace_back(pageInfo);
    clusterBuilder.CommitColumnRange(17, 0, 100, pageRange);
-   builder.AddClusterWithDetails(clusterBuilder.MoveDescriptor().Unwrap());
+   builder.AddCluster(clusterBuilder.MoveDescriptor().Unwrap());
    RClusterGroupDescriptorBuilder cgBuilder;
    RNTupleLocator cgLocator;
    cgLocator.fPosition = 1337U;
    cgLocator.fBytesOnStorage = 42;
-   cgBuilder.ClusterGroupId(256).PageListLength(137).PageListLocator(cgLocator);
-   cgBuilder.AddCluster(84);
-   builder.AddClusterGroup(std::move(cgBuilder));
+   cgBuilder.ClusterGroupId(256).PageListLength(137).PageListLocator(cgLocator).NClusters(1).EntrySpan(100);
+   std::vector<DescriptorId_t> clusterIds{84};
+   cgBuilder.AddClusters(clusterIds);
+   builder.AddClusterGroup(cgBuilder.MoveDescriptor().Unwrap());
 
    auto desc = builder.MoveDescriptor();
-   auto context = RNTupleSerializer::SerializeHeaderV1(nullptr, desc);
+   auto context = RNTupleSerializer::SerializeHeader(nullptr, desc);
    EXPECT_GT(context.GetHeaderSize(), 0);
    auto bufHeader = std::make_unique<unsigned char []>(context.GetHeaderSize());
-   context = RNTupleSerializer::SerializeHeaderV1(bufHeader.get(), desc);
+   context = RNTupleSerializer::SerializeHeader(bufHeader.get(), desc);
 
    std::vector<DescriptorId_t> physClusterIDs;
    for (const auto &c : desc.GetClusterIterable()) {
@@ -579,36 +606,41 @@ TEST(RNTuple, SerializeFooter)
    EXPECT_EQ(desc.GetNClusters(), physClusterIDs.size());
    context.MapClusterGroupId(256);
 
-   auto sizePageList = RNTupleSerializer::SerializePageListV1(nullptr, desc, physClusterIDs, context);
+   auto sizePageList = RNTupleSerializer::SerializePageList(nullptr, desc, physClusterIDs, context);
    EXPECT_GT(sizePageList, 0);
    auto bufPageList = std::make_unique<unsigned char []>(sizePageList);
-   EXPECT_EQ(sizePageList, RNTupleSerializer::SerializePageListV1(bufPageList.get(), desc, physClusterIDs, context));
+   EXPECT_EQ(sizePageList, RNTupleSerializer::SerializePageList(bufPageList.get(), desc, physClusterIDs, context));
 
-   auto sizeFooter = RNTupleSerializer::SerializeFooterV1(nullptr, desc, context);
+   auto sizeFooter = RNTupleSerializer::SerializeFooter(nullptr, desc, context);
    EXPECT_GT(sizeFooter, 0);
    auto bufFooter = std::make_unique<unsigned char []>(sizeFooter);
-   EXPECT_EQ(sizeFooter, RNTupleSerializer::SerializeFooterV1(bufFooter.get(), desc, context));
-
-   RNTupleSerializer::DeserializeHeaderV1(bufHeader.get(), context.GetHeaderSize(), builder);
-   RNTupleSerializer::DeserializeFooterV1(bufFooter.get(), sizeFooter, builder);
+   EXPECT_EQ(sizeFooter, RNTupleSerializer::SerializeFooter(bufFooter.get(), desc, context));
 
+   RNTupleSerializer::DeserializeHeader(bufHeader.get(), context.GetHeaderSize(), builder);
+   RNTupleSerializer::DeserializeFooter(bufFooter.get(), sizeFooter, builder);
    desc = builder.MoveDescriptor();
 
    EXPECT_EQ(1u, desc.GetNClusterGroups());
    const auto &clusterGroupDesc = desc.GetClusterGroupDescriptor(0);
+   EXPECT_EQ(0u, clusterGroupDesc.GetMinEntry());
+   EXPECT_EQ(100u, clusterGroupDesc.GetEntrySpan());
    EXPECT_EQ(1u, clusterGroupDesc.GetNClusters());
    EXPECT_EQ(137u, clusterGroupDesc.GetPageListLength());
    EXPECT_EQ(1337u, clusterGroupDesc.GetPageListLocator().GetPosition<std::uint64_t>());
    EXPECT_EQ(42u, clusterGroupDesc.GetPageListLocator().fBytesOnStorage);
-
-   std::vector<RClusterDescriptorBuilder> clusters = RClusterGroupDescriptorBuilder::GetClusterSummaries(desc, 0);
-   RNTupleSerializer::DeserializePageListV1(bufPageList.get(), sizePageList, clusters);
-   EXPECT_EQ(physClusterIDs.size(), clusters.size());
-   for (std::size_t i = 0; i < clusters.size(); ++i) {
-      desc.AddClusterDetails(clusters[i].MoveDescriptor().Unwrap());
-   }
-
    EXPECT_EQ(1u, desc.GetNClusters());
+   EXPECT_EQ(0u, desc.GetNActiveClusters());
+
+   RNTupleSerializer::DeserializePageList(bufPageList.get(), sizePageList, 0, desc);
+   const auto &verify = desc.GetClusterGroupDescriptor(0);
+   EXPECT_EQ(0u, verify.GetMinEntry());
+   EXPECT_EQ(100u, verify.GetEntrySpan());
+   EXPECT_EQ(1u, verify.GetNClusters());
+   EXPECT_EQ(137u, verify.GetPageListLength());
+   EXPECT_EQ(1337u, verify.GetPageListLocator().GetPosition<std::uint64_t>());
+   EXPECT_EQ(42u, verify.GetPageListLocator().fBytesOnStorage);
+
+   EXPECT_EQ(1u, desc.GetNActiveClusters());
    const auto &clusterDesc = desc.GetClusterDescriptor(0);
    EXPECT_EQ(0, clusterDesc.GetFirstEntryIndex());
    EXPECT_EQ(100, clusterDesc.GetNEntries());
@@ -644,10 +676,10 @@ TEST(RNTuple, SerializeFooterXHeader)
    builder.AddFieldLink(0, 42);
    builder.AddColumn(17, 17, 42, RColumnModel(EColumnType::kInt32, true), 0);
 
-   auto context = RNTupleSerializer::SerializeHeaderV1(nullptr, builder.GetDescriptor());
+   auto context = RNTupleSerializer::SerializeHeader(nullptr, builder.GetDescriptor());
    EXPECT_GT(context.GetHeaderSize(), 0);
    auto bufHeader = std::make_unique<unsigned char[]>(context.GetHeaderSize());
-   context = RNTupleSerializer::SerializeHeaderV1(bufHeader.get(), builder.GetDescriptor());
+   context = RNTupleSerializer::SerializeHeader(bufHeader.get(), builder.GetDescriptor());
 
    builder.BeginHeaderExtension();
    builder.AddField(RFieldDescriptorBuilder()
@@ -690,13 +722,13 @@ TEST(RNTuple, SerializeFooterXHeader)
    context.MapSchema(builder.GetDescriptor(), /*forHeaderExtension=*/true);
 
    auto desc = builder.MoveDescriptor();
-   auto sizeFooter = RNTupleSerializer::SerializeFooterV1(nullptr, desc, context);
+   auto sizeFooter = RNTupleSerializer::SerializeFooter(nullptr, desc, context);
    EXPECT_GT(sizeFooter, 0);
    auto bufFooter = std::make_unique<unsigned char[]>(sizeFooter);
-   EXPECT_EQ(sizeFooter, RNTupleSerializer::SerializeFooterV1(bufFooter.get(), desc, context));
+   EXPECT_EQ(sizeFooter, RNTupleSerializer::SerializeFooter(bufFooter.get(), desc, context));
 
-   RNTupleSerializer::DeserializeHeaderV1(bufHeader.get(), context.GetHeaderSize(), builder);
-   RNTupleSerializer::DeserializeFooterV1(bufFooter.get(), sizeFooter, builder);
+   RNTupleSerializer::DeserializeHeader(bufHeader.get(), context.GetHeaderSize(), builder);
+   RNTupleSerializer::DeserializeFooter(bufFooter.get(), sizeFooter, builder);
 
    desc = builder.MoveDescriptor();
 
